using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using System;


namespace AT{
    namespace UBF{
        namespace Arbiters{

            /** Alternates in a round-robin fashion between the voting Behaviors every X seconds.
             *
             * Once a Behavior is selected, it will remain the selected Behavior until X seconds have elapsed OR if
             * it returns a NULL action vote. At that point, this will select the next non-null vote in round-robin 
             * order. If there are NO non-null votes, it will select the same Behavior and return its null Action vote.
             * This preserves the order.
             * 
             * If only one child votes, that child's Action is chosen. 
             */
            public class AlternateEveryXSeconds : AT.UBF.Arbiter {

                public float seconds;
                int currSelection;
                int prevSelection;
                float secondsSinceLastSelection;

                public AlternateEveryXSeconds(){
                    seconds = 0.0f;
                    currSelection = 0;
                    prevSelection = 0;
                    secondsSinceLastSelection = 0.0f;
                }

                public override AT.UBF.Action selectAction(List<AT.UBF.Action> candidateActions){
                    if(candidateActions.Count < 1){
                        Debug.Log("Cannot select action from empty Actions list.");
                        return null;
                    }

                    // If the previous selection is still voting AND not enough time has elapsed, return that Action
                    if(!candidateActions[currSelection].IsNull() && (Time.time - secondsSinceLastSelection < seconds)){
                        return candidateActions[currSelection];
                    }

                    do{
                        currSelection += 1;
                        if(currSelection >= candidateActions.Count){
                            // hit the last vote in the list, start again at the beginning
                            currSelection = 0;
                        }

                        if(!candidateActions[currSelection].IsNull()){
                            prevSelection = currSelection;
                            secondsSinceLastSelection = Time.time;
                            //Debug.Log("Alternating Behavior's to : " + currSelection);
                            return candidateActions[currSelection];
                        }
                    }while(currSelection != prevSelection);

                    return candidateActions[currSelection]; 
                }
                public override void _ToDict(Dictionary<string, object> dict){
                    dict["Seconds"] = this.seconds;
                }

                public override void _FromDict(Dictionary<string, object> dict){
                    string aName = this.GetType().Name;
                    seconds = Util.FloatFromJsonDict("Seconds", seconds, aName, dict);
                }
            }   
        }
    }
}