﻿using UnityEngine;
using System.Collections;
using System.Collections.Generic;

namespace AT{
    namespace UBF{
        namespace Arbiters{
        
            /**  Combines non-conflicting Actions according to priority (higher-priority overwrite motors of lower-priority Actions).
             *
             *  The way I've chosen to implement this is to first order the Action list. This then takes the
             *  lowest priority Action as the base Action. We then iterate through the rest in increasing priority order.
             *  If the current (higher priority) action sets a given motor (e.g. driveForward) we overwrite our current
             *  base Action for that *particular* motor. This leaves all other motors as is. 
             *
             *  If Actions have the same priority, the behavior is undefined due to their nondeterministic ordering.
             */
            public class PriorityFusion : AT.UBF.Arbiter {


                public override AT.UBF.Action selectAction(List<AT.UBF.Action> candidateActions){
                    if(candidateActions.Count < 1){
                        Debug.Log("Cannot select action from empty Actions list.");
                        return null;
                    }

                    // Sort the list in place - memory efficient at the expense of CPU cycles
                    candidateActions.Sort((x, y) => x.priority.CompareTo(y.priority));

                    AT.UBF.Action resultAction = candidateActions[0]; // start with the lowest priority action as the base

                    for(int i=1; i<candidateActions.Count; i++){ // this is now iterating in priority order
                        AT.UBF.Action next = candidateActions[i];
                        resultAction.overwrite(next); // overwrite as needed
                    }

                    return resultAction;
                }
            }   
        }
    }
}
