using UnityEngine;
using System.Collections;
using System.Collections.Generic;

namespace AT{
    namespace UBF{
        namespace Arbiters{

            /** Children are considered in order with earlier children overriding later children on a per-motor basis.
             *
             *  Example: If we have two children who vote like this:
             *      Behavior    Priority    DriveForward    TurnTracksRight    TurnTurretRight    FireGun
             *      ChildA      0.9         1.0             --                 --                 fire 
             *      ChildB      0.9         0.5             --                 0.1                --
             *      Result      0.9         1.0             --                 0.1                fire
             *
             *  The result would be that ChildA, because it is first, wins the priority, forward, and fireGun motors,
             *  while ChildB wins the turret motor because A didn't vote for that.
             *
             * Subsumption is essentially a PriorityFusion Arbiter where the order of the child Behaviors determine 
             * their priority. It's a matter of preference which you want to use. 
             *
             */
            public class Subsumption : AT.UBF.Arbiter {

                public override AT.UBF.Action selectAction(List<AT.UBF.Action> candidateActions){
                    if(candidateActions.Count < 1){
                        Debug.Log("Cannot select action from empty Actions list.");
                        return null;
                    }

                    AT.UBF.Action result = candidateActions[0];

                    // Only overwrite the result Action's motor IFF the prior children did not set it
                    for(int i=1; i<candidateActions.Count; i++){
                        AT.UBF.Action tempAction = candidateActions[i];
                        if(!result.driveForward.HasValue && tempAction.driveForward.HasValue){
                            result.SetDriveForward(tempAction.driveForward.Value, tempAction.driveForwardSourceID.Value);
                        }
                        if(!result.turnTracksRight.HasValue && tempAction.turnTracksRight.HasValue){
                            result.SetTurnTracksRight(tempAction.turnTracksRight.Value, tempAction.turnTracksRightSourceID.Value);
                        }
                        if(!result.turnTurretRight.HasValue && tempAction.turnTurretRight.HasValue){
                            result.SetTurnTurretRight(tempAction.turnTurretRight.Value, tempAction.turnTurretRightSourceID.Value);
                        }
                        if(!result.fireGun.HasValue && tempAction.fireGun.HasValue){
                            result.SetFireGun(tempAction.fireGun.Value, tempAction.fireGunSourceID.Value);
                        }
                    }

                    return result; 
                }
            }   
        }
    }
}
