using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using System;


namespace AT{
    namespace UBF{
        namespace Arbiters{

            /**  Selects randomly from Actions using their assigned weights (percent chance of being selected).
             *
             * For example, a Behavior with a weight of 5 will be selected 5% of the time. 
             *
             * The total of all the weights MUST total 100. Internally, these weights are translated into an range of  
             * integers. A weight of X will result in that Behavior's Action-Vote being picked X percent of the time.
             *
             * Example:  Behavior     Weight        Assigned Integers   Selected on Rand Int of
             *           LeadFire     60 %          1-60                <= 60
             *           PureFire     30 %          61-90               <= 90
             *           LagFire      10 %          91-100              <= 100
             *
             * On each selectAction, this generates an integer between 1-100, picking the Behavior's vote which 
             * corresponds to that integer. We iterate through the weights in order, selecting the Action-Vote which 
             * has translatedWeight (post) which is greater than the random integer. In this example, if we randomly 
             * generated 65, we would start with LeadFire with a post of 60. So we move on to PureFire with a post of 
             * 90, so we select its' vote (because the random integer is less than or equal to the post value).
             * This logic is difficult to explain, but makes finding which Action to select easy and fast.
             *
             * Note only the integer weights are saved (and loaded). The translated weights which are used internally 
             * are generated from those weights. This prevents any disagreement between the two and clearly define which
             * is considered the "truth."
             *
             * Example Agent JSON using WeightedRandom:
             *
             *    {"BehaviorType": "AT.UBF.CompositeBehavior",
             *      "Arbiter": { "ArbiterType": "AT.UBF.Arbiters.WeightedRandom", "Weights": [10, 90] },
             *      "Children": [
             *          {"BehaviorType": "AT.UBF.Behaviors.PureFire" },
             *          {"BehaviorType": "AT.UBF.Behaviors.LeadFire" }
             *       ]
             *    }
             *
             * The example above will result in PureFire being selected 10% of the time and LeadFire 90% of the time.
             */
            public class WeightedRandom : AT.UBF.Arbiter {

                public List<long> weights; // raw weights given at load (e.g. 15%, 30%, 55%)
                public List<int> translatedWeights; // cutoff integer for this weight, these are ordered (e.g. 15, 45, 100)

                public override AT.UBF.Action selectAction(List<AT.UBF.Action> candidateActions){
                    if(candidateActions.Count < 1){
                        Debug.Log("Cannot select action from empty Actions list.");
                        return null;
                    }
                    if(candidateActions.Count != weights.Count){
                        Debug.Log("WeightedRandom: Number of candidate Actions and weights must match.");
                        return null;
                    }

                    int randInt = UnityEngine.Random.Range(1, 100);

                    // Iterate thru the weights from lowest to highest and return the first match (see comments above)
                    for(int i=0; i<translatedWeights.Count; i++){
                        if(randInt <= translatedWeights[i]){
                            return candidateActions[i];
                        }
                    }

                    Debug.Log("WeightedRandom: No weight matched. This should not have happened!");
                    return null;
                }

                public override void _ToDict(Dictionary<string, object> dict){
                    dict["Weights"] = weights;
                }

                public override void _FromDict(Dictionary<string, object> dict){
                    try{
                        List<object> tmpList = (List<object>) dict["Weights"];
                        weights = new List<long>();
                        foreach(object obj in tmpList){
                            weights.Add((long) obj);
                        }
                    }catch(KeyNotFoundException){
                        Debug.LogError("'Weights' key not found in WeightedRandom dictionary");
                        return;
                    }catch(InvalidCastException){
                        Debug.LogError("'Weights' value in WeightedRandom Arguments is not a valid List of integers");
                        return;
                    }

                    // First make sure weights add to 100
                    int total = 0;
                    foreach(long x in weights){
                        total += (int) x;
                    }
                    if(total != 100){
                        Debug.LogError("'Weights' key must be a List of Integers that add to 100");
                        return;
                    }

                    // Now translate the weights to our internal notation 
                    translatedWeights = new List<int>();
                    int currPost = 0;
                    foreach(long x in weights){
                        currPost += (int) x; // find our new dividing "post" (last should always be 100)
                        translatedWeights.Add(currPost);
                    }
                }

                /** Return the _ToDict key:object values as a pretty string meant for the _ToDict Dialog Box.
                 *
                 *  We override this method so we can handle the Weights carefully. Otherwise they show up as 
                 *  "Systems.Collections.Generic.List[System.Int64]" in the dialog box which isn't helpful.
                 */
                public override string GetToDictDialogBoxString(string infoStr, bool shadowTwin){
                    Dictionary<string, object> dict = new Dictionary<string, object>();
                    _ToDict(dict);

                    // This part could be taken out since we don't have any other ToDict info, but then someone might
                    // forget when they add new args...
                    infoStr += String.Format("\n{0}{1}\n", this.GetType(), (shadowTwin ? "  (Shadow Twin)" : ""));
                    foreach(KeyValuePair<string, object> entry in dict){
                        if(entry.Key == "Weights"){continue;} // we'll handle these ourselves
                        infoStr += String.Format("\t{0}: ", entry.Key);
                        infoStr += String.Format("{0}\n", entry.Value.ToString());
                    }

                    // Handle the weights 
                    infoStr += "\tWeights: [";
                    for(int i=0; i<translatedWeights.Count; i++){
                        string comma = (i+1<translatedWeights.Count) ? "," : "";
                        infoStr += String.Format("{0}{1} ", translatedWeights[i], comma); 
                    }
                    infoStr += "]\n";

                    return infoStr;
                }
            }   
        }
    }
}

