﻿using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using System;

namespace AT{
    namespace UBF{
        namespace Behaviors {
            /** IF bullets are in view, determine if they're in one of three locations:
             * current position, behind the tank, in front of the tank.  Move to which
             * ever option has the least incoming bullets.
             * 
             * Written by Stan Vernier
             *
             */
            public class DodgeBullet : AT.UBF.Behavior {
                private float buffer = 2;
                public float priority = 0.8f;

                protected override AT.UBF.Action genActionInternal(AT.UBF.State state){
					AT.UBF.Action action = new AT.UBF.Action();
                    
                    

					if(state.visibleBullets.Count > 0){
                        Vector2 myPosition = state.self.getPosition();
                        Vector2 myVelocity = state.self.getVelocity();
                        Vector2 tankBuffer = state.bodyLength.first - state.bodyLength.second;
                        
                        Vector2 futureTankPos = myPosition + myVelocity + tankBuffer/2;
                        Vector2 pastTankPos = myPosition - myVelocity - tankBuffer/2;
                        float goingForwardMultiplier = 1.0f;
                        float dirAngle = Util.getRelativeAngleToTarget(myPosition, futureTankPos);
                        float relAngle = dirAngle - state.self.getTracksRotation();

                        if(Mathf.Abs(relAngle) > 90f){
                            goingForwardMultiplier = -1.0f;
                        }

                        
                        Tuple<Vector2, Vector2> futureLine = new Tuple<Vector2, Vector2>(myPosition, futureTankPos);
                        Tuple<Vector2, Vector2> pastLine = new Tuple<Vector2, Vector2>(myPosition, pastTankPos);
                        int hitAtCurrPos = 0;
                        int hitAlongFuture = 0;
                        int hitAlongPast = 0;
                        int totalHits = 0;

                        //count up the bullet intersections
                        foreach(Tuple<Vector2, Vector2> bullet in state.visibleBullets){
                            Tuple<Vector2, Vector2> currTankLine = getMaxTankWidthLine(bullet, state.bodyLength, myPosition);
                            Vector2 bulletPos = bullet.first;
                            Vector2 bulletVel = bullet.second;
                            
                            Vector2 bulletDest = bulletPos + bulletVel * buffer;
                            Tuple<Vector2, Vector2> bulletLine = new Tuple<Vector2, Vector2>(bulletPos, bulletDest);

                            if(LineIntersection.doLinesIntersect(currTankLine, bulletLine)){
                                hitAtCurrPos++;
                            }
                            if(LineIntersection.doLinesIntersect(futureLine, bulletLine)){
                                hitAlongFuture++;
                            }
                            if(LineIntersection.doLinesIntersect(pastLine, bulletLine)){
                                hitAlongPast++;
                            }
                        }
                        totalHits += hitAtCurrPos + hitAlongFuture + hitAlongPast;
                        
                        //figure out which has the least "hits" and move in that direction
                        if (hitAtCurrPos >= hitAlongFuture && hitAtCurrPos >= hitAlongPast) {
                            if (hitAlongFuture >= hitAlongPast) {
                                action.SetDriveForward(-1.0f * goingForwardMultiplier, voteSourceID);
                            } else {
                                action.SetDriveForward(1.0f * goingForwardMultiplier, voteSourceID);
                            }
                        } else if (hitAtCurrPos >= hitAlongFuture) {
                            action.SetDriveForward(1.0f * goingForwardMultiplier, voteSourceID);
                        } else if (hitAtCurrPos >= hitAlongPast) {
                            action.SetDriveForward(-1.0f * goingForwardMultiplier, voteSourceID);
                        } else {  //hitAtCurrPos < hitAlongPast && hitAtCurrPos < hitAlongFuture
                            action.SetDriveForward(0f, voteSourceID);
                        }
                        if(totalHits > 0){
                            action.SetPriority(priority, voteSourceID);
                        }
					}

					return action;
				}

                private Tuple<Vector2, Vector2> getMaxTankWidthLine(Tuple<Vector2, Vector2> bulletLine, Tuple<Vector2, Vector2> bodyLength, Vector2 tankPos) {
                    Vector2 v = bulletLine.second;
                    Vector2 numerator = new Vector2(-v.y, v.x);

                    Vector2 tankSize = (bodyLength.first - bodyLength.second) * 1.2f;
                    float scaling = Mathf.Sqrt(Mathf.Pow(tankSize.x, 2) + Mathf.Pow(tankSize.y, 2)) / 2f;

                    Vector2 p3 = tankPos + numerator / Mathf.Sqrt(Mathf.Pow(v.x, 2) + Mathf.Pow(v.y, 2)) * scaling;
                    Vector2 p4 = tankPos + numerator / Mathf.Sqrt(Mathf.Pow(v.x, 2) + Mathf.Pow(v.y, 2)) * -scaling;
                    
                    return new Tuple<Vector2, Vector2>(p3, p4);
                }
                
                public override void _ToDict(Dictionary<string, object> dict, bool saveLastAction){
                    dict["Priority"] = priority;
                    dict["Buffer"] = buffer;
                }
                
                public override void _FromDict(Dictionary<string, object> dict){
                    string bName = this.GetType().Name;
                    priority =  Util.FloatFromJsonDict("Priority", priority, bName, dict);
                    buffer = Util.FloatFromJsonDict("Buffer", buffer, bName, dict);
                }
            }
        }
    }
}