using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using System;


namespace AT{
    namespace UBF{
        namespace Behaviors{

            /** Drives in a straight line toward the nearest wall at the desired speed until we are within the desired 
             * distance (+/- buffer) to said wall. Otherwise does not vote.
             *
             * Keep in mind the tank's size when setting the desired distance to wall and buffer values. It's very easy
             * to cause the Tank to get hung up on a wall. Optimal values may change depending on the map (wide open vs.
             * lots of walls that are close together).
             */
            public class HugClosestWall : AT.UBF.Behavior {

                public float priorityWhenNotWithinBuffer = 0.5f;
                public float desiredSpeed = 1.0f;
                public float desiredDistToWall = 1.5f; 
                public float buffer = 0.8f;

                protected override AT.UBF.Action genActionInternal(AT.UBF.State state){
                    AT.UBF.Action action = new AT.UBF.Action();

                    if(state.walls.Count < 1){return action;} // No walls to hide behind; DO NOT VOTE

                    // Start with first wall as closest, then check the rest
                    Tuple<Vector2, Vector2> closestWall = state.walls[0];

                    Tuple<float, Vector2> tmp = Util.distToLineSegment(state.self.getPosition(), closestWall.first, closestWall.second);
                    float distToClosestWall = tmp.first;
                    Vector2 closestPointInWall = tmp.second;

                    for(int i=1; i<state.walls.Count; i++){
                        Tuple<Vector2, Vector2> wall = state.walls[i];
                        tmp = Util.distToLineSegment(state.self.getPosition(), wall.first, wall.second);
                        float dist = tmp.first;
                        Vector2 point = tmp.second;

                        if(dist < distToClosestWall){
                            // We found a closer wall 
                            closestWall = wall;
                            distToClosestWall = dist;
                            closestPointInWall = point;
                        }
                    }

                    // Debug.Log("Closest Wall to our current position: " + closestWall.first + "  " + 
                    //            closestWall.second + " At Point: " + closestPointInWall);

                    // Are we within the buffer distance to this wall?
                    bool tooFarAway = (distToClosestWall > desiredDistToWall + buffer);
                    bool tooClose   = (distToClosestWall < desiredDistToWall - buffer);

                    if(!tooFarAway && !tooClose){ return action; } // Within buffer; DO NOT VOTE

                    action.SetTurnTracksRight(
                        Util.findShortestTrackRotationToTarget(state.self.getPosition(), closestPointInWall, state.self.getTracksRotation()),
                        voteSourceID);
                    action.SetPriority(priorityWhenNotWithinBuffer, voteSourceID);   

                    if(tooFarAway){ // drive straight toward the wall
                        action.SetDriveForward(desiredSpeed, voteSourceID);
                    }else{ // tooClose; drive straight back from the wall
                        action.SetDriveForward(0.0f - desiredSpeed, voteSourceID); 
                    }
                        
                    return action;
                }

                public override void _ToDict(Dictionary<string, object> dict, bool saveLastAction){
                    dict["PriorityWhenNotWithinBuffer"] = priorityWhenNotWithinBuffer;
                    dict["DesiredSpeed"] = desiredSpeed;
                    dict["DesiredDistToWall"] = desiredDistToWall;
                    dict["Buffer"] = buffer; 
                }

                /** Loads arguments from a dictionary, verifying as needed.
                 */
                public override void _FromDict(Dictionary<string, object> dict){
                    string bName = this.GetType().Name;
                    priorityWhenNotWithinBuffer =  Util.FloatFromJsonDict("PriorityWhenNotWithinBuffer", priorityWhenNotWithinBuffer, bName, dict);
                    desiredSpeed =  Util.FloatFromJsonDict("DesiredSpeed", desiredSpeed, bName, dict);
                    desiredDistToWall =  Util.FloatFromJsonDict("DesiredDistToWall", desiredDistToWall, bName, dict);
                    buffer =  Util.FloatFromJsonDict("Buffer", buffer, bName, dict);
                }
            }
        }
    }
}