using UnityEngine;
using System.Collections;
using System.Collections.Generic;

namespace AT{
    namespace UBF{
        namespace Behaviors{

            /** Shoots at an enemy tank with a leading shot (using estimated time of of flight and enemy velocity).
             *
             *  This does NOT vote if no enemy is in sight. When an enemy IS in sight, it tracks the first one in the 
             *  list. 
             * 
             *  When the tracked enemy is outside of our firing zone, only vote on the turret rotation. The firing zone 
             *  is set via argument, and represents the maximum turret turn rotation that we can vote for and still fire 
             *  at the enemy. For example, a firing zone of 0.1 means we only fire when we vote to turn the turret at 
             *  10% of full power.
             */
            public class LeadFire : AT.UBF.Behavior {

                public float priorityWhenEnemyWithinFiringZone = 1.0f;
                public float priorityWhenEnemyOutsideFiringZone = 1.0f;
                public float firingZonePercentTurretRotation = 0.1f;

                public float bulletSpeedFudgeFactor = 2.0f; // The bullet's mass is quite small and without this, this 
                // math seems to think it'll take longer than it does to reach the player resulting in misses

                protected override AT.UBF.Action genActionInternal(AT.UBF.State state){
                    AT.UBF.Action action = new AT.UBF.Action();

                    if(state.visibleEnemyTanks.Count < 1){
                        return action; // Do not vote if there isn't anything to shoot
                    }

                    TankState enemy = state.visibleEnemyTanks[0];
                    Vector2 enemyPosition = enemy.getPosition();
                    Vector2 enemyVelocity = enemy.getVelocity();

                    // Time of flight to reach enemy
                    float t = (enemyPosition - state.self.getPosition()).magnitude / AT.GameManager.instance.bulletSpeed;
                    t = t / bulletSpeedFudgeFactor;
                    
                    // Project future position
                    Vector2 futurePos = enemyPosition + (enemyVelocity * t);
                    //Debug.Log("FuturePos: " + futurePos);

                    // Try to match the desired lead
                    action.SetTurnTurretRight(
                        Util.findShortestTurretRotationToTarget(state.self.getPosition(), futurePos, state.self.getTurretRotation()),
                        voteSourceID);
                    
                    if(Mathf.Abs(action.turnTurretRight.Value) < firingZonePercentTurretRotation){
                        // close enough to desired lead to shoot
                        action.SetFireGun(true, voteSourceID);
                        action.SetPriority(priorityWhenEnemyWithinFiringZone, voteSourceID);
                    }else{
                        action.SetPriority(priorityWhenEnemyOutsideFiringZone, voteSourceID);
                    }

                    return action;
                }

                /** Save arguments to dictionary.
                 */
                public override void _ToDict(Dictionary<string, object> dict, bool saveLastAction){
                    dict["PriorityWhenEnemyWithinFiringZone"] = priorityWhenEnemyWithinFiringZone;
                    dict["PriorityWhenEnemyOutsideFiringZone"] = priorityWhenEnemyOutsideFiringZone;
                    dict["FiringZonePercentTurretRotation"] = firingZonePercentTurretRotation;
                }

                /** Loads arguments from a dictionary, verifying as needed.
                 */
                public override void _FromDict(Dictionary<string, object> dict){
                    string bName = this.GetType().Name;
                    priorityWhenEnemyWithinFiringZone = Util.FloatFromJsonDict("PriorityWhenEnemyWithinFiringZone", priorityWhenEnemyWithinFiringZone, bName, dict);
                    priorityWhenEnemyOutsideFiringZone = Util.FloatFromJsonDict("PriorityWhenEnemyOutsideFiringZone", priorityWhenEnemyOutsideFiringZone, bName, dict);
                    firingZonePercentTurretRotation = Util.FloatFromJsonDict("FiringZonePercentTurretRotation", firingZonePercentTurretRotation, bName, dict);
                }
            }
        }
    }
}

