using UnityEngine;
using System.Collections;
using System.Collections.Generic;

namespace AT{
    namespace UBF{
        namespace Behaviors{

            /** Shoots directly at an enemy tank with no lead or lag (pure pursuit aka aim at their current position, 
             * ignoring their velocity).
             *
             *  This does NOT vote if no enemy is in sight. When an enemy IS in sight, it tracks the first one in the 
             *  list. 
             * 
             *  When the tracked enemy is outside of our firing zone, only vote on the turret rotation. The firing zone 
             *  is set via argument, and represents the maximum turret turn rotation that we can vote for and still fire 
             *  at the enemy. For example, a firing zone of 0.1 means we only fire when we vote to turn the turret at 
             *  10% of full power.
             */
            public class PureFire : AT.UBF.Behavior {

                public float priorityWhenEnemyWithinFiringZone = 1.0f;
                public float priorityWhenEnemyOutsideFiringZone = 1.0f;
                public float firingZonePercentTurretRotation = 0.1f;


                protected override AT.UBF.Action genActionInternal(AT.UBF.State state){
                    AT.UBF.Action action = new AT.UBF.Action();

                    if(state.visibleEnemyTanks.Count < 1){
                        return action; // Do not vote if there isn't anything to shoot
                    }

                    Vector2 enemyPosition = state.visibleEnemyTanks[0].getPosition();

                    // Try to aim directly at the enemy's CURRENT position
                    action.SetTurnTurretRight(
                        Util.findShortestTurretRotationToTarget(state.self.getPosition(), enemyPosition, state.self.getTurretRotation()),
                        voteSourceID);
                    
                    if(Mathf.Abs(action.turnTurretRight.Value) < firingZonePercentTurretRotation){
                        // close enough to desired aim position to shoot
                        action.SetFireGun(true, voteSourceID);    
                        action.SetPriority(priorityWhenEnemyWithinFiringZone, voteSourceID);
                    }else{
                        action.SetPriority(priorityWhenEnemyOutsideFiringZone, voteSourceID);
                    }

                    return action;
                }

                /** Save arguments to dictionary.
                 */
                public override void _ToDict(Dictionary<string, object> dict, bool saveLastAction){
                    dict["PriorityWhenEnemyWithinFiringZone"] = priorityWhenEnemyWithinFiringZone;
                    dict["PriorityWhenEnemyOutsideFiringZone"] = priorityWhenEnemyOutsideFiringZone;
                    dict["FiringZonePercentTurretRotation"] = firingZonePercentTurretRotation;
                }

                /** Loads arguments from a dictionary, verifying as needed.
                 */
                public override void _FromDict(Dictionary<string, object> dict){
                    string bName = this.GetType().Name;
                    priorityWhenEnemyWithinFiringZone = Util.FloatFromJsonDict("PriorityWhenEnemyWithinFiringZone", priorityWhenEnemyWithinFiringZone, bName, dict);
                    priorityWhenEnemyOutsideFiringZone = Util.FloatFromJsonDict("PriorityWhenEnemyOutsideFiringZone", priorityWhenEnemyOutsideFiringZone, bName, dict);
                    firingZonePercentTurretRotation = Util.FloatFromJsonDict("FiringZonePercentTurretRotation", firingZonePercentTurretRotation, bName, dict);
                }
            }
        }
    }
}

