﻿using System;
using UnityEngine;
using UnityEditor;
using System.Collections;
using System.Collections.Generic;

namespace AT{
    namespace UBF{ 
        /**  Simple action object which does not dictate min, max or any other limitations on value.
         *
         *  Limits on these attributes are left to the Controller class to handle!
         *
         *  Since the primitives are not nullable types, we wrap them with Nullable. You can use
         *  foo.HasValue to test if it has been set and foo.Value for the actual value (if it has one).
         *
         *  Is serializable for persistence - making unit tests possible.
         */
        [System.Serializable]
        public class Action {
            /** 0.0 for none, 1.0 for highest priority */
            public float priority;
            public int   prioritySourceID; // TODO: Is this needed or useful?
            
            public Nullable<float> driveForward;
            public Nullable<int>   driveForwardSourceID;
            public Nullable<float> turnTracksRight;
            public Nullable<int>   turnTracksRightSourceID;
            public Nullable<float> turnTurretRight;
            public Nullable<int>   turnTurretRightSourceID;
            public Nullable<bool>  fireGun;
            public Nullable<int>   fireGunSourceID;

            // GUI Layout Constants (for Inspector GUI, etc.)
            public static int colWidth = 50;
            public static int rowHeight = 14;

            public void SetPriority(float newPriority, int voteSourceID){
                priority = newPriority;
                prioritySourceID = voteSourceID;
            }

            public void SetDriveForward(float newDriveForward, int voteSourceID){
                driveForward = newDriveForward;
                driveForwardSourceID = voteSourceID;
            }

            public void SetTurnTracksRight(float newTurnTracksRight, int voteSourceID){
                turnTracksRight = newTurnTracksRight;
                turnTracksRightSourceID = voteSourceID;
            }

            public void SetTurnTurretRight(float newTurnTurretRight, int voteSourceID){
                turnTurretRight = newTurnTurretRight;
                turnTurretRightSourceID = voteSourceID;
            }

            public void SetFireGun(bool newFireGun, int voteSourceID){
                fireGun = newFireGun;
                fireGunSourceID = voteSourceID;
            }

            /** Create and return a deep clone of this Action.
             *
             * Currently, Action is made up of only basic/value types, so we can safely use the overwrite() method to
             * perform the clone. If a reference type is ever added to Action, this will need to be changed. For more
             * info on the difference, refer to https://msdn.microsoft.com/en-us/library/4d43ts61(v=vs.90).aspx
             */
            public Action Clone(){
                Action clone = new Action();    
                clone.overwrite(this); 
                return clone;
            }

            /** Does this vote have NULL for each and every motor? (Behavior abstained.)
             *
             * This ignores priority.
             */
            public bool IsNull(){
                if(driveForward.HasValue){
                    return false;
                }else if(turnTracksRight.HasValue){
                    return false;
                }else if(turnTurretRight.HasValue){
                    return false;
                }else if(fireGun.HasValue){
                    return false;
                }
                return true;
            }

            /** Overwrite our attributes with those defined in Action a (unless they are null, then keep our value).
             *
             *  Used for some Fusion arbitration methods.
             *
             * TODO: Move this logic into the fusion arbiters that use it (e.g. PriorityFusion)? 
             */
            public void overwrite(AT.UBF.Action a){
                this.priority = a.priority;
                this.prioritySourceID = a.prioritySourceID;

                if(a.driveForward.HasValue){
                    this.driveForward = a.driveForward.Value;
                    this.driveForwardSourceID = a.driveForwardSourceID;
                }
                if(a.turnTracksRight.HasValue){
                    this.turnTracksRight = a.turnTracksRight.Value;
                    this.turnTracksRightSourceID = a.turnTracksRightSourceID;
                }
                if(a.turnTurretRight.HasValue){
                    this.turnTurretRight = a.turnTurretRight.Value;
                    this.turnTurretRightSourceID = a.turnTurretRightSourceID;
                }
                if(a.fireGun.HasValue){
                    this.fireGun = a.fireGun.Value;
                    this.fireGunSourceID = a.fireGunSourceID;
                }
            }

            /** Draw an Inspector GUI representing this Action.
             *
             * bool showRealAction:  
             *     If real, show headers. If a Shadow Action, don't show headers and add an '(S)' at the end of the 
             *      motor values to denote it as such.
             *
             * AT.UBF.Action parentLastAction:
             *     If not null, make any motor/priority values that our parent selected bold. This allows us to quickly 
             *     see which votes are winning and which aren't at all levels of the tree. If null, nothing is bold.
             */
            public void DrawInspectorGUI(bool showRealAction, AT.UBF.Action parentLastAction){
                AT.UBF.Action la;
                if(parentLastAction == null){
                    la = new Action();
                }else{
                    la = parentLastAction;
                }

                if(showRealAction){
                    GUILayout.BeginHorizontal();
                    GUILayout.Label("Priority", 
                        prioritySourceID != 0 && la.prioritySourceID == prioritySourceID ?  EditorStyles.boldLabel : EditorStyles.label,
                        GUILayout.Width(colWidth), GUILayout.Height(rowHeight));
                    GUILayout.Label("Forward", 
                        (la.driveForward.HasValue && (la.driveForwardSourceID == driveForwardSourceID)) ? EditorStyles.boldLabel : EditorStyles.label,
                        GUILayout.Width(colWidth), GUILayout.Height(rowHeight));
                    GUILayout.Label("Tracks", 
                        (la.turnTracksRight.HasValue && (la.turnTracksRightSourceID == turnTracksRightSourceID)) ? EditorStyles.boldLabel : EditorStyles.label,
                        GUILayout.Width(colWidth), GUILayout.Height(rowHeight));
                    GUILayout.Label("Turret", 
                        (la.turnTurretRight.HasValue && (la.turnTurretRightSourceID == turnTurretRightSourceID)) ? EditorStyles.boldLabel : EditorStyles.label,
                        GUILayout.Width(colWidth), GUILayout.Height(rowHeight));
                    GUILayout.Label("FireGun", 
                        (la.fireGun.HasValue && (la.fireGunSourceID == fireGunSourceID)) ? EditorStyles.boldLabel : EditorStyles.label,
                        GUILayout.Width(colWidth), GUILayout.Height(rowHeight));
                    GUILayout.EndHorizontal();  
                }

                GUILayout.BeginHorizontal();
                GUILayout.Label(String.Format("{0:F2}", priority), 
                    prioritySourceID != 0 && la.prioritySourceID == prioritySourceID ?  EditorStyles.boldLabel : EditorStyles.label,
                    GUILayout.Width(colWidth), GUILayout.Height(rowHeight));
                GUILayout.Label(driveForward.HasValue ? String.Format("{0:F2}", driveForward.Value) : "--", 
                    (la.driveForward.HasValue && (la.driveForwardSourceID == driveForwardSourceID)) ? EditorStyles.boldLabel : EditorStyles.label,
                    GUILayout.Width(colWidth), GUILayout.Height(rowHeight));
                GUILayout.Label(turnTracksRight.HasValue ? String.Format("{0:F2}", turnTracksRight.Value) : "--", 
                    (la.turnTracksRight.HasValue && (la.turnTracksRightSourceID == turnTracksRightSourceID)) ? EditorStyles.boldLabel : EditorStyles.label,
                    GUILayout.Width(colWidth), GUILayout.Height(rowHeight));
                GUILayout.Label(turnTurretRight.HasValue ? String.Format("{0:F2}", turnTurretRight.Value) : "--", 
                    (la.turnTurretRight.HasValue && (la.turnTurretRightSourceID == turnTurretRightSourceID)) ? EditorStyles.boldLabel : EditorStyles.label,
                    GUILayout.Width(colWidth), GUILayout.Height(rowHeight));
                GUILayout.Label(fireGun.HasValue ? String.Format("{0}", fireGun.Value) : "--", 
                    (la.fireGun.HasValue && (la.fireGunSourceID == fireGunSourceID)) ? EditorStyles.boldLabel : EditorStyles.label,
                    GUILayout.Width(colWidth), GUILayout.Height(rowHeight));
                if(!showRealAction){
                    GUILayout.Label("(S)", GUILayout.Width(colWidth), GUILayout.Height(rowHeight));
                }
                GUILayout.EndHorizontal(); 
            }

            public virtual Dictionary<string, object> ToDict(){
                Dictionary<string, object> dict = new Dictionary<string, object>();

                dict["Priority"] = priority;
                dict["DriveForward"] = driveForward;
                dict["TurnTracksRight"] = turnTracksRight;
                dict["TurnTurretRight"] = turnTurretRight;
                dict["FireGun"] = fireGun;

                return dict;
            }
        }
    }   
}
