﻿using UnityEngine;
using UnityEditor;
using System.Collections;
using System.Collections.Generic;
using System;
using System.IO;

namespace AT{
    namespace UBF{
        public class Behavior{

            [HideInInspector]
            public AT.UBF.Action lastAction;
            public int voteSourceID;
            protected bool toggleGUI = false; // Start with the Tree Folded
            [HideInInspector]
            public AT.UBF.Behavior parent;
            [HideInInspector]
            public AT.UBF.Behavior shadowTwin;

            // GUI Globals for the UBF Tree view 
            public static int yMargin = 40;
            public static int xMargin = 5;

            /** Returns Action to be executed by the agent or considered by a CompositeBehavior.
             *
             * Must return an Action - even if it is null/empty.
             */ 
            public AT.UBF.Action genAction(AT.UBF.State state){
                Action newAction = genActionInternal(state);
                // This preserves the Action's results in case the reference we pass is changed or overwritten by 
                // Arbiters, etc. 
                lastAction = newAction.Clone(); 
                return newAction;  
            }

            protected virtual AT.UBF.Action genActionInternal(AT.UBF.State state){
                return new AT.UBF.Action(); 
            }


            /** Do these two UBF Behavior trees have exactly the same structure?
             *
             *  Identical structure means the trees must have exactly the same number of Behaviors and in the 
             *  same layout (parent-child edges). This does not compare exact Behavior and Arbiter types, beyond 
             *  verifying that CompositeBehavior's are matched to CompositeBehaviors and plain Behaviors to plain
             *  Behaviors. 
             *
             *  For some examples, see below (CB ~ CompositeBehavior, X-Z are plain Behaviors):
             *
             *  Example 1:
             *
             *       CB      CB
             *      /  \    /  \    These are considered to have Identical Structure, even though the leaf Behaviors
             *      X  Y    Z  Z    are different.
             *
             *  Example 2:
             *
             *       CB       CB
             *      /  \    / | \    These do NOT have Identical Structure, because one has three leaf Behaviors while
             *      X  Y    X Y Z    the other only has two. 
             *
             *  Example 3:
             *
             *       CB        CB
             *      /  \      /  \     These DO have Identical Structure.
             *      X  CB     Z  CB     
             *        /  \      /  \
             *        Y  Z      X   X
             */
            public virtual bool IdenticalStructure(AT.UBF.Behavior b, Dictionary<int, AT.UBF.Behavior> twins){
                if(b is AT.UBF.Behavior){
                    twins[voteSourceID] = b;
                    return true;
                }
                return false;
            }

            /** Save the reference to our Shadow UBF twin using a voteSourceID -> AT.UBF.Behavior lookup Dictionary.
             *
             *  This assumes this is a safe operation, meaning the two Trees ARE identical. 
             *  The lookup Dictionary will most likely come from IdenticalStructure().  
             */
            public virtual void SetShadowTwin(Dictionary<int, AT.UBF.Behavior> twins){
                shadowTwin = twins[voteSourceID];
            }

            
            /** 
             */
            public Dictionary<string, object> ToDict(bool saveLastAction){
                Dictionary<string, object> dict = new Dictionary<string, object>();
                dict["BehaviorType"] = this.GetType().FullName; // Required 
                if(saveLastAction){
                    // Save lastAction (check for null first)
                    dict["LastAction"] = lastAction == null ? null : lastAction.ToDict();
                }
                this._ToDict(dict, saveLastAction);
                return dict;
            }

            public virtual void _ToDict(Dictionary<string, object> dict, bool saveLastAction){
                // dict["TurningSpeed"] = this.turningSpeed; // Example
            }
            
            /** Creates a new Behavior instance from the provided Dictionary.
             *
             * Any parsing errors, invalid or missing argument errors are caught here and NULL is 
             * returned. The error will be logged, but we do this because these errors are expected.
             */
            public static AT.UBF.Behavior FromDict(Dictionary<string, object> dict){
                // Step 1: Find the fully specified Behavior Class Name using the '_type' key
                string behaviorClassName;
                try{
                    behaviorClassName = dict["BehaviorType"] as string;
                }catch(KeyNotFoundException){
                    Debug.LogError("Did not find a 'BehaviorType' key specifying the Behavior Class name.");
                    return null; 
                }

                // Step 2: Create a new instance of that Behavior (cannot create instance of MonoBehavior derived classes)
                AT.UBF.Behavior tempBehavior = (AT.UBF.Behavior) Activator.CreateInstance(null, behaviorClassName).Unwrap();
                if(tempBehavior == null){ 
                    Debug.LogError("Could not find the Behavior with Class name '" + behaviorClassName + "'");
                    return null; 
                }

                // Step 3: Setup vote source ID for identifying which Behavior/Arbiter a specific vote came from
                tempBehavior.voteSourceID = Controller.VoteSourceID; // Controller auto-increments this ID

                // Step 4: Load any options, arguments and internal state to the Behavior
                try{
                    tempBehavior._FromDict(dict);
                    return tempBehavior;
                }catch(Exception e){
                    Debug.LogError("Failed to load Behavior '"+behaviorClassName+"' from Dictionary. " + e.ToString());
                    return null;
                }
            }

            public virtual void _FromDict(Dictionary<string, object> dict){
                // this.turningSpeed = (float) dict["TurningSpeed"]; // Example
            }

            /** Show any information provided by _ToDict (i.e. internal state and arguments) via Editor Dialog box.
             */
            public virtual void ShowToDictDialogBox(){
                Dictionary<string, object> dict = new Dictionary<string, object>();
                _ToDict(dict, false);

                string infoStr = String.Format("{0}\n", this.GetType());
                foreach(KeyValuePair<string, object> entry in dict){
                    infoStr += String.Format("\t{0}: ", entry.Key);
                    infoStr += String.Format("{0}\n", entry.Value.ToString());
                }

                if(shadowTwin != null){
                    dict = new Dictionary<string, object>();
                    shadowTwin._ToDict(dict, false);

                    infoStr += String.Format("\n{0}  (Shadow Twin)\n", this.GetType());
                    foreach(KeyValuePair<string, object> entry in dict){
                        infoStr += String.Format("\t{0}: ", entry.Key);
                        infoStr += String.Format("{0}\n", entry.Value.ToString());
                    }
                }

                string title = "_ToDict()  Response";
                EditorUtility.DisplayDialog(title, infoStr, "Ok");
            }

           
           /** Draw tree based Inspector view allowing the developer to see this Behaviors name and last action returned
            */
            public virtual void DrawInspectorGUI(AT.UBF.Action parentLastAction){
                GUILayout.BeginHorizontal();
                GUILayout.BeginVertical();
                toggleGUI = EditorGUILayout.Foldout(toggleGUI, this.GetType().Name); // toggle (show/hide) foldout
                if(toggleGUI){
                    // We of course returned lastAction, but we want to highlight any parts (motors) of that Action 
                    // that were chosen by our parent (parentLastAction)
                    lastAction.DrawInspectorGUI(true, parentLastAction);
                }
                GUILayout.EndVertical();
                GUILayout.EndHorizontal();
            }

            /** Get the number of leaf (terminal) Behaviors.
             *
             * This exists to allow for GUI layouts (it tells us how much room a given tree will take on screen) along
             * with the known width and height requirements for each Behavior (and corresponding lastAction object).
             */
            public virtual int GetNumLeafBehaviors(){
                return 1; // THIS behavior MUST be a leaf (it's not a CompositeBehavior)
            }

            /** Treating this Behavior as the root of the tree, get this tree's height.
             *
             * "The height of a node is the length of the longest downward path to a leaf from that node. The height 
             *  of the root is the height of the tree."
             *   -- Wikipedia  https://en.wikipedia.org/wiki/Tree_(data_structure)#Terminology
             *
             * This exists to allow for GUI layouts (it tells us how much room a given tree will take on screen) along
             * with the known width and height requirements for each Behavior (and corresponding lastAction object).
             */
            public virtual int GetHeight(){
                return 1; // there are no leaves below us, so we have a height of 1
            }

            /** Get the width that our Inspector GUI will require.  
             */
            public int GetUiWidth(){
                return 6 * AT.UBF.Action.colWidth; // 6 columns
            }

            /** Get the height that our Inspector GUI will require.  
             *
             *  This will change depending on whether or not we have a shadow twin to show.
             */
            public int GetUiHeight(){
                return AT.UBF.Action.rowHeight * (shadowTwin != null ? 5 : 4);
            }

            /** Get the width that our Inspector GUI will require, including margin. 
             */
            public int GetUiWPlusMargin(){
                return GetUiWidth() + xMargin;
            }

            /** Get the height that our Inspector GUI will require, including margin. 
             */
            public int GetUiHPlusMargin(){
                return GetUiHeight() + yMargin;
            }

            public virtual void DrawAdvInspectorGUI(Rect drawRect){
                GUI.Window(voteSourceID, 
                    new Rect(drawRect.center.x - GetUiWidth() / 2, drawRect.yMin,  GetUiWidth(), GetUiHeight()), 
                    DrawAdvInspectorWindow, 
                    this.GetType().Name);
            }

            public virtual void DrawAdvInspectorWindow(int window){
                lastAction.DrawInspectorGUI(true, parent == null ? null : parent.lastAction);
                if(shadowTwin != null){
                    shadowTwin.lastAction.DrawInspectorGUI(false, 
                        (shadowTwin.parent == null ? null : shadowTwin.parent.lastAction));
                }

                // When the user clicks this Window, show this Behavior's _ToDict() output
                if((Event.current.button == 0) && (Event.current.type == EventType.MouseUp)){
                    ShowToDictDialogBox();
                }
            }
        }
    }
}
