﻿using UnityEngine;
using UnityEditor;
using System.Collections;
using System.Collections.Generic;
using System;

namespace AT{
    namespace UBF{
        public class CompositeBehavior : AT.UBF.Behavior {

            private AT.UBF.Arbiter arbiter = null;
            private List<AT.UBF.Behavior> children = new List<AT.UBF.Behavior>();

            

            public void setArbiter(AT.UBF.Arbiter arbiter){
                this.arbiter = arbiter;
            }

            public void addChild(AT.UBF.Behavior behavior){
                /*  Adds a Behavior (can be another CompositeBehavior) which will then be used in genAction according to the current
                 *  arbiter's policy.
                 */
                behavior.parent = this;
                this.children.Add(behavior);
            }



            protected override Action genActionInternal(AT.UBF.State state){
                /* Controller may or may not call this every tick. It depends on how the tree is constructed and any conditional logic.
                 *
                 * Must return an Action.
                 */
                if(this.children.Count < 1){
                    Debug.Log("CompositeBehavior must have at least one child Behavior to generate a valid Action!");
                    return null;
                }
                if(this.arbiter == null){
                    Debug.Log("CompositeBehavior must have an Arbiter to select from the candidate Actions!");
                    return null;
                }

                List<AT.UBF.Action> candidateActions = new List<AT.UBF.Action>();
                for(int i=0; i<this.children.Count; i++){
                    candidateActions.Add(this.children[i].genAction(state));
                }
                return this.arbiter.selectAction(candidateActions);
            }

            /** Used to return Arguments dictionary for Save().
             *
             * CompositeBehavior requires three Arguments: Arbiter, Children and Arguments.
             * The last one (Arguments) are for the Arbiter. I realize this might be confusing :(
             */
            public override void _ToDict(Dictionary<string, object> dict, bool saveLastAction){
                dict["Arbiter"] = arbiter.ToDict(); 
                
                // Save each child (CompositeBehavior children make this recursive)
                List<Dictionary<string, object>> tempChildren = new List<Dictionary<string, object>>();
                for(int i=0; i<this.children.Count; i++){
                    tempChildren.Add(this.children[i].ToDict(saveLastAction));
                }
                dict["Children"] = tempChildren;
            }

            /** Loads arguments from a dictionary, verifying as needed.
             */
            public override void _FromDict(Dictionary<string, object> dict){
                AT.UBF.Arbiter tempArbiter = AT.UBF.Arbiter.FromDict((dict["Arbiter"] as Dictionary<string, object>));
                if(tempArbiter == null){
                    Debug.LogError("Failed to load Arbiter.");
                    throw new Exception(); // TODO, is this how we should do this error?
                } 
                setArbiter(tempArbiter);

                
                List<object> tempChildren;
                try{
                    tempChildren = dict["Children"] as List<object>; 
                }catch(KeyNotFoundException ){
                    Debug.LogError("Invalid CompositeBehavior. Did not have a 'Children' key and corresponding Array of Behaviors.");
                    throw new Exception(); // TODO, is this how we should do this error?
                }

                // Now load each Child Behavior (which could be another CompositeBehavior) and add it to this
                for(int i=0; i<tempChildren.Count; i++){
                    Dictionary<string, object> child = tempChildren[i] as Dictionary<string, object>;
                    AT.UBF.Behavior tempBehavior = AT.UBF.Behavior.FromDict(child); // recursive
                    if(tempBehavior == null){
                        Debug.LogError("Failed to load a child Behavior.");
                        throw new Exception(); // TODO, is this how we should do this error?
                    }
                    addChild(tempBehavior); 
                }
            }

            /** Show any information provided by the Arbiters _ToDict (i.e. internal state and arguments) via Editor Dialog box.
             */
            public override void ShowToDictDialogBox(){
                string infoStr = "";
                infoStr = arbiter.GetToDictDialogBoxString(infoStr, false);

                if(shadowTwin != null){
                    AT.UBF.Arbiter sArb = ((AT.UBF.CompositeBehavior) shadowTwin).arbiter;
                    infoStr = sArb.GetToDictDialogBoxString(infoStr, true);
                }

                string title = "_ToDict()  Response";
                EditorUtility.DisplayDialog(title, infoStr, "Ok");
            }

            public override void DrawInspectorGUI(AT.UBF.Action parentLastAction){
                GUILayout.BeginHorizontal();
                GUILayout.BeginVertical();
                toggleGUI = EditorGUILayout.Foldout(toggleGUI, this.GetType().Name + " --  Arbiter: " + arbiter.GetType().Name);
                if(toggleGUI){
                    // We of course returned lastAction, but we want to highlight any parts (motors) of that Action 
                    // that were chosen by our parent (parentLastAction)
                    lastAction.DrawInspectorGUI(true, parentLastAction);

                    GUILayout.BeginHorizontal();
                    GUILayout.Space(30); // indent all children
                    GUILayout.BeginVertical();
                    for(int i=0; i<this.children.Count; i++){
                        // Now draw our children Behaviors and past the Action WE returned (not our parents' lastAction)
                        this.children[i].DrawInspectorGUI(lastAction);
                    }
                    GUILayout.EndVertical();
                    GUILayout.EndHorizontal();
                }

                GUILayout.EndVertical();
                GUILayout.EndHorizontal();
            }

            public override void DrawAdvInspectorGUI(Rect drawRect){
                // Draw this CompositeNode centered above its' children
                GUI.Window(voteSourceID, 
                    new Rect(drawRect.center.x - GetUiWidth() / 2, 
                             drawRect.yMin,  
                             GetUiWidth(), 
                             GetUiHeight()), 
                    DrawAdvInspectorWindow, 
                    this.GetType().Name + " --  " + arbiter.GetType().Name);

                // Draw each child Behavior
                float sumW =  drawRect.xMin;
                for(int i=0; i<children.Count; i++){
                    float w = GetUiWPlusMargin() * children[i].GetNumLeafBehaviors();
                    float y = drawRect.yMin + GetUiHPlusMargin();
                    children[i].DrawAdvInspectorGUI(new Rect(sumW, y, w, 100));

                    // Show Parent/Child relationship using Handles 
                    Handles.BeginGUI();
                    Handles.color = Color.black;
                    Handles.DrawLine(new Vector2(drawRect.center.x, drawRect.yMin + GetUiHeight()), new Vector2(sumW + w / 2, y));
                    Handles.EndGUI();

                    sumW += w;
                }
            }

            /** Get the number of leaf (terminal) Behaviors.
             *
             * This exists to allow for GUI layouts (it tells us how much room a given tree will take on screen) along
             * with the known width and height requirements for each Behavior (and corresponding lastAction object).
             */
            public override int GetNumLeafBehaviors(){
                int total = 0;
                if(children.Count == 0){
                    return 1;  // No children, so this is a terminal Behavior
                }
                // Otherwise, get the number of Terminal Behaviors by asking our Children
                for(int i=0; i<this.children.Count; i++){
                    total += children[i].GetNumLeafBehaviors();
                }
                return total;
            }

            /** Treating this Behavior as the root of the tree, get this tree's height.
             *
             * "The height of a node is the length of the longest downward path to a leaf from that node. The height 
             *  of the root is the height of the tree."
             *   -- Wikipedia  https://en.wikipedia.org/wiki/Tree_(data_structure)#Terminology
             *
             * This exists to allow for GUI layouts (it tells us how much room a given tree will take on screen) along
             * with the known width and height requirements for each Behavior (and corresponding lastAction object).
             */
            public override int GetHeight(){
                if(children.Count == 0){
                    return 1;  // No children, so this is a terminal Behavior (height of 1)
                }

                // Otherwise, get the height by asking our Children
                int maxHeight = 0;
                int tempHeight;
                for(int i=0; i<this.children.Count; i++){
                    tempHeight = children[i].GetHeight();
                    if(tempHeight > maxHeight){
                        maxHeight = tempHeight;
                    }
                }
                return maxHeight + 1; // add one to include the CompositeBehavior
            }

            /** Do these two UBF Behavior trees have exactly the same structure?
             *
             *  Overrides the base method to check that:
             *    1. Behavior 'b' is a CompositeBehavior (like us)
             *    2. Behavior 'b' has the same number of children that we do
             *    3. Recursively check each of our Children using their IdenticalStructure(b) method
             *
             *  See Behavior.IdenticalStructure() for more info on how this works.
             */
            public override bool IdenticalStructure(AT.UBF.Behavior b, Dictionary<int, AT.UBF.Behavior> twins){
                // Step 1: Is Behavior 'b' is a CompositeBehavior (like us)
                if(!(b is AT.UBF.CompositeBehavior)){return false;}
                AT.UBF.CompositeBehavior cb = (AT.UBF.CompositeBehavior) b;

                // Step 2: Does Behavior 'b' have the same number of children as us?
                if(children.Count != cb.children.Count){return false;}

                // Step 3: Recursively check each of our Children using their IdenticalStructure(b) method
                for(int i=0; i<this.children.Count; i++){
                    if(!this.children[i].IdenticalStructure(cb.children[i], twins)){
                        return false;
                    }
                }

                // Step 4: Everything looks good
                twins[voteSourceID] = b;
                return true;
            }

            /** Save the reference to our Shadow UBF twin using a voteSourceID -> AT.UBF.Behavior lookup Dictionary.
             *
             *  This assumes this is a safe operation, meaning the two Trees ARE identical. 
             *  The lookup Dictionary will most likely come from IdenticalStructure().  
             */
            public override void SetShadowTwin(Dictionary<int, AT.UBF.Behavior> twins){
                shadowTwin = twins[voteSourceID];

                for(int i=0; i<this.children.Count; i++){
                    children[i].SetShadowTwin(twins);
                }
            }
        }
    }
}
