﻿using UnityEngine;
using UnityEditor;
using System.Collections;
using System.Collections.Generic;


namespace AT{
    namespace UBF{ 
        /** Base Tank UBF Controller. Handles execution of the UBF Behaviors and resulting Action.
         */
        public class Controller : AT.Tank {

            [HideInInspector]
            [SerializeField]
            private static int voteSourceID = 0;
            public static int VoteSourceID{
                get{
                    voteSourceID += 1; // Increment to the next ID as needed
                    return voteSourceID;
                }
                private set{
                    voteSourceID = value;
                }
            }

            [HideInInspector]
            [SerializeField] 
            private string agentJSONPath;
            public string AgentJSONPath{ 
                get{return agentJSONPath;}
                set{
                    Debug.Log("Setting JSON path: " + value);
                    // only take the filename (no directories) because we ASSUME it's in the Agents directory
                    // This forces the file paths to be portable between developers (instead of C:/Joe/AutoTank/Foo)
                    agentJSONPath = "Agents/" + System.IO.Path.GetFileName(value); 
                    SetName(agentJSONPath.Replace("/", "-").Replace("\\", "-"));
                }
            }

            [HideInInspector]
            [SerializeField] 
            private string shadowAgentJSONPath;
            public string ShadowAgentJSONPath{
                get{return shadowAgentJSONPath;}
                set{
                    if(value == null){
                        shadowAgentJSONPath = null; 
                        return;
                    }
                    Debug.Log("Setting Shadow JSON path: " + value);
                    // only take the filename (no directories) because we ASSUME it's in the Agents directory
                    // This forces the file paths to be portable between developers (instead of C:/Joe/AutoTank/Foo)
                    shadowAgentJSONPath = "Agents/" + System.IO.Path.GetFileName(value);
                }
            }

            // Are the real and shadow UBF trees identical in structure?
            [HideInInspector]
            public bool shadowAgentIdentical;

            [HideInInspector]
            public AT.UBF.Behavior behavior;       // Real UBF Tree
            [HideInInspector]
            public AT.UBF.Behavior shadowBehavior; // Shadow UBF Tree -- used for comparing output to the real UBF tree
            [HideInInspector]
            public AT.UBF.State state; 
            [HideInInspector]
            public AT.UBF.Action currAction; /** Last generated Action. Used by 3rd parties to passively inspect UBF's current operation. */
        

            public override void Awake(){
                base.Awake(); // MUST be done first
                if(behavior == null){ // skip if we already loaded the UBF tree (e.g. using FromDict() )
                    loadBehaviorTree();
                    if(behavior == null){Util.QuitImmediately();} // we cannot recover from this error
                }
                this.state = new AT.UBF.State(this.gameObject, vision);
            }

            /** Update is called once per update - updating the state, generating an action 
             * and then executing that action.
             */
            public void Update(){
                UpdateState();
                currAction = GenerateAction();
                this.ExecuteAction(currAction);
            }

            /** Update base (Tank) and UBF specific state.
             *
             * This must be performed before generating an Action using the UBF behavior tree.
             *
             * Note: This has been factored out to make testing easier.
             */
            public override void UpdateState(){
                base.UpdateState();
                if(state == null){
                    Debug.LogError("Controller.state is NULL!"); // can we just initialize it here if it's null?
                    Util.QuitImmediately();
                    return;
                }
                state.Update();// 'this' isn't passed in since Unity provides methods to get (child) components as needed
            }

            /** Generate a new Action using the UBF tree and current State.
             *
             * Note: This has been factored out to make testing easier.
             */
            public AT.UBF.Action GenerateAction(){
                if(behavior == null){
                    Debug.LogError("Controller.behavior is NULL!");
                    Util.QuitImmediately();
                    return null;
                }
                AT.UBF.Action action = behavior.genAction(state);
                if(shadowBehavior != null){
                    shadowBehavior.genAction(state);
                }
                return action;
            }

            /**  Accepts an action and then executes any non-NULL attributes of that action.
             */     
            protected void ExecuteAction(AT.UBF.Action a){
                AT.UBF.Action action = (AT.UBF.Action) a;
                // All values _must_ be specified, so assume 0.0 is default if none given
                if(!action.turnTracksRight.HasValue){action.turnTracksRight = 0.0f;}
                if(!action.driveForward.HasValue){action.driveForward = 0.0f;}
                if(!action.turnTurretRight.HasValue){action.turnTurretRight = 0.0f;}

                // Now execute modified Action
                Vector3 tracks_turned = TurnTracks(action.turnTracksRight.Value);
                DriveForward(action.driveForward.Value);
                RotateTurret(action.turnTurretRight.Value, tracks_turned);  
                if(action.fireGun.HasValue && action.fireGun.Value){
                    FireGun();
                }
            }

            /** Attempts to load the persisted Agent from a JSON file, looking in the Agents/ directory by default.
             * 
             * Renames GameObject after the JSON filename.
             *
             * Factored out to use in OnValidate and Awake()
             */
            public void loadBehaviorTree(){
                Debug.Log("Loading " + AgentJSONPath);
                if(AgentJSONPath.Length <= 0){return;}

                Dictionary<string, object> tempJsonDict = Util.JsonDictFromFile(AgentJSONPath);
                if(tempJsonDict == null){return;} // failed to load from file

                AT.UBF.Behavior tempBehavior = AT.UBF.Behavior.FromDict(tempJsonDict);
                if(tempBehavior == null){return;} // failed to load due to errors

                behavior = tempBehavior;
                // Set GameObject name to match to more easily find it in the editor
                SetName(AgentJSONPath.Replace("/", "-").Replace("\\", "-")); 

                // Try to load the Shadow UBF Tree (ONLY if the real UBF tree loaded above)
                if(ShadowAgentJSONPath.Length <= 0){return;}

                tempBehavior = AT.UBF.Behavior.FromDict(Util.JsonDictFromFile(ShadowAgentJSONPath));
                if(tempBehavior == null){return;} // failed to load due to errors

                shadowBehavior = tempBehavior;

                // Check to see if these are identical in structure and if so, record the shadow twin lookup (dict)
                Dictionary<int, AT.UBF.Behavior> shadowTwins = new Dictionary<int, AT.UBF.Behavior>();
                shadowAgentIdentical = behavior.IdenticalStructure(shadowBehavior, shadowTwins);
                Debug.Log("Loaded Shadow UBF Tree is " + (shadowAgentIdentical ? "identical" : "not identical"));
                if(shadowAgentIdentical){
                    // Trees are identical in structure, so set ALL of their shadow twins for a full comparison
                    behavior.SetShadowTwin(shadowTwins);     
                }else{
                    // Only set the top/root node's shadow twin, because we cannot otherwise compare these two trees
                    behavior.shadowTwin = shadowBehavior;
                }         
            }

            /** Used to persist internal state dictionary for Save()
             *
             * saveOnlyJsonPath: Will save only the agent's JSON path for loading it's UBF tree. Setting to false will
             *   save the entire UBF tree (using Behaviors' ToDict()). By using only the JSON path, you can save space 
             *   and change the UBF tree in one place (the JSON file) versus each save. This is especially important 
             *   when breaking changes are made to a Behavior's ToDict/Fromdict. 
             *
             * saveLastAction: Will save the last Action-Vote generated by this UBF tree. Note FromDict does NOT use 
             *   this information. It's only meant for external analysis. Instead, FromDict causes the UBF tree to 
             *   re-vote and thus regenerate lastAction.
             *
             */
            public Dictionary<string, object> ToDict(bool saveOnlyJsonPath, bool saveLastAction){
                Dictionary<string, object> dict = base.ToDict(); // save transform info for tracks and turret

                dict["_type"] = this.GetType().FullName; // mostly just an FYI 

                if(saveOnlyJsonPath){
                    dict["agentJSONPath"] = AgentJSONPath;
                    if(saveLastAction){ // Save only TOP MOST lastAction
                        dict["LastAction"] = behavior.lastAction.ToDict();
                    }
                }else{
                    // will save ENTIRE behavior tree and EACH Behavior's lastAction
                    dict["behavior"] = behavior.ToDict(saveLastAction); 
                }

                return dict;
            }

            /** Loads internal state from a dictionary, verifying as needed.
             *
             * If the "agentJSONPath" key is present, load the UBF tree from file. If not, try loading it from the 
             * "behavior" key (note these are different data structures).
             */
            public override void FromDict(Dictionary<string, object> dict){
                base.FromDict(dict);

                if(dict.ContainsKey("agentJSONPath")){
                    AgentJSONPath = dict["agentJSONPath"] as string; 
                    loadBehaviorTree();
                }else if(dict.ContainsKey("behavior")){    
                    behavior = AT.UBF.Behavior.FromDict((dict["behavior"] as Dictionary<string, object>));
                    SetName("No File"); // TODO: Name this better (normally we use the JSON file name)
                }else{
                    Debug.LogError("Agent Dictionary had neither a JSON file path, nor entire UBF tree to load.");
                }   
            } 
        }
    }   
}
