using UnityEngine;
using System;
using System.Collections;
using UnityEditor;

namespace AT {
    namespace UBF {
        /** Provides custom Inspector for the Controller. 
         *
         * This creates a UBF Tree view when the game is running. This allows you to traverse the 
         * tree and view its operation in real time.
         */
        [CustomEditor(typeof(Controller))]
        public class ControllerEditor : Editor {

            bool showStateFoldout = false;

            public override void OnInspectorGUI(){
                Controller controller = (Controller) target;
                DrawDefaultInspector();

                GUILayout.BeginHorizontal();
                GUILayout.Label("Current Agent Path:", EditorStyles.boldLabel);
                GUILayout.Label(controller.AgentJSONPath);
                GUILayout.EndHorizontal();

                GUI.enabled = false;
                GUILayout.TextArea("This assumes (and enforces) all Agent paths are relative and are in the Agents directory.");
                GUI.enabled = true;

                // Don't allow agent load during play (though Controller CAN do this, it should only be done when 
                // loading from a file, etc.)
                if(Application.isPlaying){ GUI.enabled = false; } 
                if(GUILayout.Button("Load")){
                       LoadAgentDialog();
                }
                GUI.enabled = true;

                GUILayout.BeginHorizontal();
                GUILayout.Label("Shadow Agent Path:", EditorStyles.boldLabel);
                GUILayout.Label(controller.ShadowAgentJSONPath);
                GUILayout.EndHorizontal();

                // Don't allow agent load during play (though Controller CAN do this, it should only be done when 
                // loading from a file, etc.)
                if(Application.isPlaying){ GUI.enabled = false; } 
                if(GUILayout.Button("Load Shadow Agent")){LoadShadowAgentDialog();}
                if(GUILayout.Button("Clear Shadow Agent")){controller.ShadowAgentJSONPath = null;EditorUtility.SetDirty(target);}
                GUI.enabled = true;

                // If not GameManager instance exists yet OR it has Debug Aids disabled: stop drawing here
                if(!AT.GameManager.instance || AT.GameManager.instance.disableDebugAids){return;}

                showStateFoldout = EditorGUILayout.Foldout(showStateFoldout, "UBF State");
                if(showStateFoldout){ 
                    GUILayout.BeginHorizontal();
                    GUILayout.Space(30);
                    GUILayout.BeginVertical();
                    if(Application.isPlaying){
                        try{
                            controller.state.DrawInspectorGUI();
                        }catch(NullReferenceException){
                            GUILayout.Label("State Not Set");
                        }
                    }else{
                        GUILayout.Label("Not Running");
                    }
                    GUILayout.EndVertical();
                    GUILayout.EndHorizontal();
                }


                GUILayout.Label("UBF Tree:", EditorStyles.boldLabel);
                if(Application.isPlaying){ 
                    // Force the Editor to repaint if the game is being played
                    // this allows the dev to see their tree executing in real-time
                    EditorUtility.SetDirty(target);
                    try{
                        controller.behavior.DrawInspectorGUI(controller.currAction);
                    }catch(NullReferenceException){
                        GUILayout.Label("Behavior Tree Not Set");
                    }
                }else{
                    GUILayout.Label("Not Running");
                }
            }

            /** Load an Agent's UBF tree from JSON file.
             */
            private void LoadAgentDialog(){
                Controller controller = (Controller) target;
                string path = EditorUtility.OpenFilePanel(
                                "Choose Agent JSON File",
                                "Agents/",
                                "json");
                if (path.Length != 0) {
                    controller.AgentJSONPath = path;
                    EditorUtility.SetDirty(target); // make sure changed attributes are preserved on the object
                }
            }

            
            /** Load an Agent's Shadow UBF tree from JSON file.
             */
            private void LoadShadowAgentDialog(){
                Controller controller = (Controller) target;
                string path = EditorUtility.OpenFilePanel(
                                "Choose Shadow Agent JSON File",
                                "Agents/",
                                "json");
                if (path.Length != 0) {
                    controller.ShadowAgentJSONPath = path;
                    EditorUtility.SetDirty(target); // make sure changed attributes are preserved on the object
                }
            }
        }
    }
}