using UnityEditor;
using UnityEngine;
using UnityEngine.UI;
using System;
using System.IO;
using System.Collections;
using System.Collections.Generic;

namespace AT {

    /** Singleton to manage the game, including setup and persistence (hold references, settings, score, etc).
     *
     *
     * Allows the game to be saved periodically, paused and for "time shifting" or loading these old saves. This allows
     * the developer to go back in time to inspect behavior, but also to (as long as the game is still paused) go 
     * forward in this progression of saves. Because the developer can change the save interval mid-game, this stream
     * is not necessarily evenly spaced in time! 
     *
     * This manager also allows the entire set of saves to be saved to disc as JSON upon the game ending. This is useful
     * for inspecting old playthroughs.
     *
     * Note: Singletons are generally frowned upon (because they are essentially global state) but we need a canonical 
     * game referee, persistence manager and repository for global settings that should affect the entire game. Unity
     * prefabs works well, but more than once I've not realized that a prefab setting change hadn't propagated to all
     * of its instances. 
     */
    public class GameManager : MonoBehaviour {

        public static GameManager instance; // Static instance allows it to be accessed by any script

        // Disables UBF tree view, time rewind/forward and unit/integration test tools (useful for performance boost?)
        public bool disableDebugAids = false; 

        public GameObject tankPrefab;
        public GameObject bulletPrefab;
        public Texture2D cursorTexture;

        public float bulletSpeed = 7.0f;
        public float bulletSecondsToLive = 4.0f; 
		public int gameLengthInSeconds = 60;

        [HideInInspector]
        [SerializeField]
        private string tankLayer = "TankLayer";
        public string TankLayer{
            get{return tankLayer;}
            private set{tankLayer = value;} 
        }

        [HideInInspector]
        [SerializeField]
        private string bulletLayer = "BulletLayer"; // On their own layer to allow for Vision RayCasting 
        public string BulletLayer{
            get{return bulletLayer;}
            private set{bulletLayer = value;}
        }

        [HideInInspector]
        [SerializeField]
        private string wallLayer = "WallLayer";
        public string WallLayer{
            get{return wallLayer;}
            private set{wallLayer = value;}
        }

        // Saves / Time Shifting
        [HideInInspector]
        public int currSave;
        [HideInInspector]
        public List<Dictionary<string, object>> saves = new List<Dictionary<string, object>>();
        [HideInInspector]
        public float nextSaveTime;
        
        [SerializeField]
        private float saveInterval = 1.0f;
        [HideInInspector]
        public float SaveInterval{
            get{return saveInterval;}
            set{saveInterval = Mathf.Round(value*10)/10;} // round to nearest tenth
        }
        [HideInInspector]
        public float minSaveInterval = 0.1F;
        [HideInInspector]
        public float maxSaveInterval = 10.0F;
        public bool recordPlaythroughsToDisc = false; // save all saves to disc as JSON by default
        public bool saveOnlyAgentJsonPath = true; // load Agent's from their JSON files, (false saves the entire UBF tree)
        // save the UBF's last Action-Vote to JSON file. This is never used when loading and is only meant for external 
        // analysis. 
        public bool saveLastAction = false;

        // Default list of colors to help identify tanks during play (displayed on Tank's sprite AND in UBF.State)
        public Tuple<string, Color>[] tankColors = {
                new Tuple<string, Color>("blue", Color.blue),
                new Tuple<string, Color>("cyan", Color.cyan),
                new Tuple<string, Color>("gray", Color.gray),
                new Tuple<string, Color>("green", Color.green),
                new Tuple<string, Color>("magenta", Color.magenta),
                new Tuple<string, Color>("red", Color.red),
                new Tuple<string, Color>("white", Color.white),
                new Tuple<string, Color>("yellow", Color.yellow)
            };
        

        /** Maintains Singleton instance and starts the game.
         *
         * If you use this Singleton, it may not exist during 'Awake()'. If you need to perform one-
         * time setup, try using Start() instead.
         *
         * Adapted from https://unity3d.com/learn/tutorials/projects/2d-roguelike-tutorial/writing-game-manager
         */
        void Awake(){ 
            if (instance == null){ // Check if instance already exists. If not, set instance to this
                Debug.Log("GameManager starting...");
                instance = this;
                DontDestroyOnLoad(this);
            }else if (instance != this){ // If instance already exists and it's not this:
                //Then destroy this. This enforces our singleton pattern, meaning there can only ever be one instance of a GameManager.
                Destroy(gameObject); 
            }   
            
            StartGame();
        }

        /** Auto-save the game if necessary and then check for one of our game over conditions.
         *
         * The game ends if...
         *     1. There is only one tank left (tank wins; score shown)
         *     2. There are no tanks left (user forgot to add tanks to the scene or the tanks simultaneously died)
         *     3. Two or more remaining tanks which are all on the same team, but not "team" None (team wins; scores are shown)
         *     4. Time limit is exceeded (only scores are shown)
         */
        public void FixedUpdate(){
            SaveGame(false); // only check to see if we need to auto save

			GameObject[] remainingTanks = GameObject.FindGameObjectsWithTag("Tank");

			if(remainingTanks.Length == 1){
				EndGame(remainingTanks[0].name + " won!\n" + getScores());
			}else if(remainingTanks.Length == 0){
				// Either they forgot to add Tanks to the scene OR both tanks blew up in the same tick
                EndGame("No tanks remaining?!");
			}else if(Time.time >= gameLengthInSeconds){
				//End the game if we've exceeded time
				EndGame("Time limit reached.\n" + getScores());
			}else{
                // Test all remaining tanks to see if they are on the same team, calling EndGame if they are
                AT.Team teamA = remainingTanks[0].GetComponent<AT.Tank>().team;
                // 'None' is special, and does not allow a "team" win
                if(teamA == AT.Team.None){return;}
                bool allRemainingTanksOnSameTeam = true;
                for(int i=1; i<remainingTanks.Length; i++){
                    if(teamA != remainingTanks[i].GetComponent<AT.Tank>().team){
                        allRemainingTanksOnSameTeam = false;
                        break; 
                    }
                }
                if(allRemainingTanksOnSameTeam){
                    EndGame("Team " + teamA.ToString() + " won!\n" + getScores());
                }
            }
	    }

        /** Pause the game, save as needed and display a game over message with scores/winners. Then calls PostGame().
         */
        public void EndGame(string gameOverMsg){
            if(TimePlaying()){ // prevents multiple calls AND winning while paused
                PauseTime();  
                // Save as necessary
                SaveGame(true); // force a save to capture very last game state
                if(recordPlaythroughsToDisc){
                    SaveAllGameSavesToDisc("Playthroughs/" + Util.getSortableDateTimeString() + ".json");
                }
                // Tell the user what happened
                Debug.Log(gameOverMsg);
                EditorUtility.DisplayDialog("Game Over", gameOverMsg, "Ok");
                // Call custom End Game logic
                PostGame();
            }
        }

        /** Called after EndGame is complete. Derive from GameManager and override this method to write custom
         * Post Game logic, such as loading a new scene or asking the player to replay the scene if they did not win.
         *
         * To load a new scene, use Application.LoadLevel("SomeScene");
         */
        public virtual void PostGame(){
            EditorApplication.ExecuteMenuItem("Edit/Play");
        }

        /** Return scores (i.e. health) of remaining tanks as a String.
         *
         * In the future, this will likely want to account for damage done to enemy tanks (add to score) and perhaps 
         * friendly fire (subtract from score).
         */
		public String getScores() {
			Tank[] tanks = GameObject.FindObjectsOfType<Tank>();
			String outString = "Remaining tanks' scores:\n";
			for(int i = 0; i < tanks.Length; i++) {
				Tank tempTank = tanks[i];
				outString += tempTank.name + " = " + tempTank.getHealth() + "\n";
			}
			return outString;
		}

        /** Pause if game is running, or Resume if game is paused.
         */
        public void PauseOrResume(){ 
            // Originally this was a one line ternary, but it skipped certain checks (as in ResumeTime) :\
            if(TimePlaying()){ 
                PauseTime(); 
            }else{ 
                ResumeTime();
            }
        }

        /** Is Unity playing and paused?
         */
        public bool TimePaused(){
            return (Application.isPlaying && EditorApplication.isPaused);
        }

        /** Is Unity playing and not paused?
         */
        public bool TimePlaying(){
            return (Application.isPlaying && !EditorApplication.isPaused);
        }

        /** Pause the Unity Editor.
         */
        public void PauseTime(){
            EditorApplication.isPlaying = false;;
        }

        /** Resume the Unity Editor.
         */
        public void ResumeTime(){
            if(GameObject.FindGameObjectsWithTag("Tank").Length < 2){
                Debug.Log("Cannot resume game with only one tank in the scene (game over).");
                // Otherwise, the game will just pause again when FixedUpdate detects the same thing
                return;
            }

            EditorApplication.isPlaying = true;
        }

        /** Called right before StartGame(). Used to setup the Agents, Map, etc.
         * 
         * The default SetupGame() does nothing. Derive from GameManager and override this method if you need custom 
         * logic. For example, if I want the user (i.e. PlayerControlled) to fight 4 rounds, each against a different 
         * agent and/or map, use this method to set those up. You just need to keep track of which round is is and 
         * remember NOT to reset the scene!
         */
        public virtual void  SetupGame(){
        }

        /** Resets the scene, loading in the map and agents as needed.
         */
        public void StartGame(){
            PauseTime(); // prevent input/agents running while we setup
            SetupGame();
            nextSaveTime = -1; // Force first AutoSave
            saves = new List<Dictionary<string, object>>();
            currSave = 0;
            Cursor.SetCursor(cursorTexture, Vector2.zero, CursorMode.Auto); // change cursor to crosshair
            SetupTankColors();
            ResumeTime();
        }

        /** Assign each tank in the scene a unique color from tankColors member.
         */
        private void SetupTankColors(){
            GameObject[] tanks = GameObject.FindGameObjectsWithTag("Tank");

            for(int i=0; i<tanks.Length; i++){
                // find the tank body and turret's sprite renderer
                SpriteRenderer[] tempSRs = tanks[i].GetComponentsInChildren<SpriteRenderer>() as SpriteRenderer[];
                foreach(SpriteRenderer sr in tempSRs){
                    sr.color = tankColors[i].second; // assign each tank a unique color based on the index
                    // This requires there be enough colors to pick from!
                }

                // find the base AT.Tank instance
                AT.Tank tankInst = tanks[i].GetComponent<AT.Tank>();
                tankInst.tankColor = tankColors[i].first; 
            }    
        }

        /** Check to see if we need to perform an auto save of the game state (based on saveInterval) OR force a save.
         */
        private void SaveGame(bool forceSave){
            if(Application.isPlaying){
                if(forceSave || nextSaveTime == -1 || Time.time > nextSaveTime){ // we just hit play OR saveInterval has elapsed
                    nextSaveTime = Time.time + saveInterval;
                    // BUG: If we don't serialize with MiniJSON, the expected type conversions aren't performed (int -> long; float -> double)
                    // and so lots of FromDict() code that expects those conversions break!
                    Dictionary<string, object> newSave = MiniJSON.Json.Deserialize(MiniJSON.Json.Serialize(ToDict())) as Dictionary<string, object>;

                    if(this.currSave < saves.Count-1){
                        // Saves need to be deleted first, because the user backed up and then played the scene from an old save
                        // To keep the save linear, we delete all saves in front of the "current" save
                        int from = this.currSave + 1;
                        int num = saves.Count - (this.currSave + 1);
                        Debug.Log("Deleting " + num.ToString() + " save(s) starting at save number " + (from+1).ToString() + " (user rewound time).");
                        saves.RemoveRange(from, num); // List.RemoveRange( index, count )
                    }

                    saves.Add(newSave);
                    this.currSave = saves.Count - 1;
                }
            }else{
                nextSaveTime = -1;
            }
        }

        /** Saves ALL of the game saves to disc as JSON. 
         *
         * Useful for recording the game for later analysis.
         *
         * WARNING: This only saves the saves List AS IS. This means that if the SaveInterval, 
         *    saveOnlyAgentJsonPath, or SaveLastAction options were changed during this playthrough, this will have some
         *    saves that follow the old settings!
         */
        public void SaveAllGameSavesToDisc(string path){
            if(saves.Count < 1){
                Debug.Log("There are no saves to persist to disc!");
                return;
            }
            PauseTime(); // Always PAUSE, we don't want a race condition between saving and the game still running

            Debug.Log("Saving ALL game saves to disc: " + path);
            // UGLY HACK: Unfortunately, MiniJSON does not support saving list even though that is valid JSON
            // so we have to put our List in a dummy Dictionary first :(
            Dictionary<string, object> tmp = new Dictionary<string, object>();
            tmp["Saves"] = saves;
            tmp = AlterAllGameSavesBeforeSaveToDisc(tmp); // optional method to filter/append data if required
            Util.JsonDictToFile(path, tmp);
        }

        /** Optional function to alter the All Game Saves before SaveAllGameSavesToDisc writes the dictionary to disc.
         *
         *  This is most useful for filtering the data or appending additional meta data. For example, I need to ask the
         *  user what strategy they used in this game save (run). This is virtual to allow GameManagers derived from 
         *  this one to override this method and do whatever they need to do.
         */
        protected virtual Dictionary<string, object> AlterAllGameSavesBeforeSaveToDisc(Dictionary<string, object> dict){
            return dict; // DO NOTHING
        }

        /** Loads ALL of the game saves (a playthrough) from disc.
         */
        public void LoadAllGameSavesFromDisc(string path){
            if(!Application.isPlaying){
                Debug.Log("Cannot load a playthrough with the game not running. (Which is confusing, I know.)");
                // reason being that we need various game objects to be initialized 
                return;
            }
            PauseTime(); // Always PAUSE, because we want to view an OLD game playthrough (don't overwrite saves)

            Debug.Log("Loading ALL game saves from disc: " + path);
            // See note about the HACK I had to use for serializing this Dict in SaveAllGameSavesToDisc
            Dictionary<string, object> tmp = Util.JsonDictFromFile(path) as Dictionary<string, object>;
            // Casting this List has been throwing compiler errors, so I had to break it apart
            List<object> tmpObjList = (List<object>) tmp["Saves"];
            saves = new List<Dictionary<string, object>>();
            foreach(Dictionary<string, object> save in tmpObjList){
                saves.Add(save);
            }

            currSave = 0; // start from the beginning
            LoadCurrentSave(); 
        }

        /** Saves the current game state to disc.
         *
         * This is different from the current save (this.saves). That List is for saving and loading entire 
         * playthroughs. This is for saving this exact moment's game state to disc. 
         */
        public void SaveCurrentGameToDisc(string path){
            PauseTime(); // Always PAUSE, we don't want a race condition between saving and the game still running
            Debug.Log("Saving current game state to disc: " + path);
            Util.JsonDictToFile(path, ToDict());
        }

        /** Load game state from disc (i.e. only one moment, not an entire playthrough).
         */
        public void LoadGameFromDisc(string path){
            if(!Application.isPlaying){
                Debug.Log("Cannot load a game moment with the game not running. (Which is confusing, I know.)");
                // reason being that we need various game objects to be initialized 
                return;
            }
            PauseTime(); // Always PAUSE
            Debug.Log("Loading game state from disc: " + path);
            Dictionary<string, object> save = Util.JsonDictFromFile(path) as Dictionary<string, object>;
            FromDict(save);
            saves = new List<Dictionary<string, object>>();
            currSave = 0;
        }

        /** Is there a previous save available to load?
         */
        public bool PrevSaveAvailable(){
            return (currSave > 0) && (saves.Count > 0);
        }

        /** Is there a save available to load after the current one?
         */
        public bool NextSaveAvailable(){
            return (currSave < saves.Count-1) && (saves.Count > 0);
        }

        /** Load the current game save into the current scene (currSave).
         *
         * See FromDict() for more information.
         */
        private void LoadCurrentSave(){
            FromDict(saves[currSave]);
        }

        /** Load the previous game save into the current scene.
         *
         * Relies on FromDict internally.
         */
        public void LoadPrevSave(){
            if(PrevSaveAvailable()){
                currSave -= 1;
                LoadCurrentSave();
            }
        }

        /** Load the next game save into the current scene.
         *
         * Relies on FromDict internally.
         */
        public void LoadNextSave(){
            if(NextSaveAvailable()){
                currSave += 1;
                LoadCurrentSave();
            }
        }

        /** Deletes all Tanks, Bullets and Walls from the scene immediately.
         */ 
        public void ClearScene(){
            // Tanks
            GameObject[] agents = GameObject.FindGameObjectsWithTag("Tank");
            for(int i=0; i < agents.Length; i++){
                UnityEngine.Object.DestroyImmediate(agents[i]); 
            }

            // Bullets 
            GameObject[] go = GameObject.FindGameObjectsWithTag("Bullet");
            for(int i=0; i < go.Length; i++){
                UnityEngine.Object.DestroyImmediate(go[i]); 
            }

            // Walls
            MapManager.instance.ClearMap(); 
        }

        /** Iterate through each AT.UBF.Testing.UnitTest in the current scene and call any and all RunUnitTest methods they have.
         */
        public void RunUnitTests(){
            AT.UBF.Testing.UnitTest[] goArray = GameObject.FindObjectsOfType<AT.UBF.Testing.UnitTest>() as AT.UBF.Testing.UnitTest[];
            Debug.Log("Running Unit tests on " + goArray.Length + " GameObjects.");
            foreach(AT.UBF.Testing.UnitTest go in goArray){
                go.RunTest();
            }
        }

        /** Iterate through each GameObject in the current scene and call any and all RunIntegrationTest methods they 
         *  have.
         */
        public void RunIntegrationTests(){
            AT.UBF.Testing.IntegrationTest[] goArray = GameObject.FindObjectsOfType<AT.UBF.Testing.IntegrationTest>() as AT.UBF.Testing.IntegrationTest[];
            Debug.Log("Running Integration tests on " + goArray.Length + " GameObjects.");
            foreach(AT.UBF.Testing.IntegrationTest go in goArray){
                go.RunTest();
            }
        }

        /** Force each UBF Agent to update its' State object and current action.
         *
         * Normally the UBF Agent(s) will do this themselves, but when the developer is loading saves from disk or 
         * memory and hasn't resumed the game, we need to force the Agents to update so they can see what they are doing
         * in the inspector. 
         */
        public void UpdateUBFAgentStateAndAction(){
            AT.UBF.Controller[] agents = GameObject.FindObjectsOfType<AT.UBF.Controller>() as AT.UBF.Controller[];
            foreach(AT.UBF.Controller agent in agents){
                agent.Update();
            }
            foreach(AT.UBF.Controller agent in agents){
                Debug.Log("Forcing agents to update State and Action");
                agent.UpdateState();
                agent.GenerateAction();
            }
        }

        /** Load a single Tank from disc (as JSON, typically from the Agents directory) to the current scene.
         *
         * Use X and Y to set initial position.
         */
        public void LoadTankFromDisc(string path, float x, float y){
            GameObject tank = Instantiate(tankPrefab); 

            AT.UBF.Controller controller = tank.GetComponent<AT.UBF.Controller>() as AT.UBF.Controller;
            controller.AgentJSONPath = path;
            controller.loadBehaviorTree();

            tank.transform.position = new Vector2(x, y);
        }

        /** Saves current game state to a Dictionary which is JSON compatible (see MiniJSON).
         *
         * This only saves the details necessary to recreate the game state, which means we assume lots of details.
         */
        public Dictionary<string, object> ToDict(){
            Dictionary<string, object> dict = new Dictionary<string, object>();
            dict["_Type"] = this.GetType().FullName; // just as an FYI

            dict["MapFilePath"] = MapManager.instance.CurrMapFilePath;

            // Save the SaveInterval (in seconds as a float) only for FYI purposes. This is NOT loaded in FromDict()
            // The user is likely to change this as they play through, so we don't want to override that setting
            // when they are rewinding/playing the game over and over which would cause unwanted problems.
            dict["SaveInterval"] = saveInterval; 

            // Find all Tanks still in the scene (i.e. still alive) and save them to a List of Dictionaries
            GameObject[] agents = GameObject.FindGameObjectsWithTag("Tank") as GameObject[];
            List<Dictionary<string, object>> tankDicts = new List<Dictionary<string, object>>(); 
            foreach(GameObject agent in agents){
                AT.UBF.Controller controller = agent.GetComponent<AT.UBF.Controller>() as AT.UBF.Controller;
                tankDicts.Add(controller.ToDict(saveOnlyAgentJsonPath, saveLastAction));
            }
            dict["Agents"] = tankDicts;

            // Find and save all Bullets 
            List<Dictionary<string, object>> bulletDicts = new List<Dictionary<string, object>>();
            GameObject[] bullets = GameObject.FindGameObjectsWithTag("Bullet") as GameObject[];
            foreach(GameObject bullet in bullets){
                Bullet bulletScript = bullet.GetComponent<Bullet>() as AT.Bullet;
                bulletDicts.Add(bulletScript.ToDict());
            }
            dict["Bullets"] = bulletDicts;

            return dict;
        }

        /** Pause time, clear the scene and load previously saved game state from a Dictionary which is JSON compatible 
         * (see MiniJSON).
         *
         * This uses only the details necessary to recreate the game state, which means we assume lots of details.
         *
         * Resets tank colors and the next save time.
         */
        public void FromDict(Dictionary<string, object> dict){
            // Step 1: Pause time, otherwise odd things can happens
            PauseTime();

            // Step 2: Clear the current scene (tanks, bullets and walls)
            ClearScene();

            // Step 3: Load the map from file
            string mapPath = dict["MapFilePath"] as string;
            MapManager.instance.CurrMapFilePath = mapPath;

            // Step 4: For each Tank in the save, create a Tank prefab and load it with those values
            List<object> tempList = ((List<object>) dict["Agents"]);
            foreach(Dictionary<string, object> tempObj in tempList){
                Dictionary<string, object> agentDict = tempObj as Dictionary<string, object>;
                GameObject tank = Instantiate(tankPrefab);
                AT.UBF.Controller controller = tank.GetComponent<AT.UBF.Controller>() as AT.UBF.Controller;
                controller.FromDict(agentDict);
            }

            // Step 5: Restore Bullets using prefab 
            tempList = ((List<object>) dict["Bullets"]);
            foreach(Dictionary<string, object> tempObj in tempList){
                Dictionary<string, object> bulletDict = tempObj as Dictionary<string, object>;
                GameObject bullet = Instantiate(bulletPrefab);
                AT.Bullet bulletScript = bullet.GetComponent<AT.Bullet>() as AT.Bullet;
                bulletScript.FromDict(bulletDict);
            }

            // Force them to update so the inspector shows current UBF tree, current action, and current state
            UpdateUBFAgentStateAndAction(); 

            SetupTankColors(); // are these going to be the same?
            nextSaveTime = Time.time + saveInterval; // reset save time 
        }

        public void SavePlaythroughFile(){
            string path = EditorUtility.SaveFilePanel(
                        "Save Playthrough File",
                        "Playthroughs/",
                        Util.getSortableDateTimeString()+".json", // default file name
                        "json");
            if (path.Length != 0) {
                SaveAllGameSavesToDisc(path);
            }
        }

        public void LoadPlaythroughFile(){
            string path = EditorUtility.OpenFilePanel(
                        "Choose Playthrough File",
                        "Playthroughs/",
                        "json");
            if (path.Length != 0) {
                LoadAllGameSavesFromDisc(path);
            }
        }

        public void SaveGameMomentFile(){
            string path = EditorUtility.SaveFilePanel(
                        "Save Game Moment File",
                        "Moments/",
                        Util.getSortableDateTimeString()+".json", // default file name
                        "json");
            if (path.Length != 0) {
                SaveCurrentGameToDisc(path);
            }
        }

        public void LoadGameMomentFile(){            
            string path = EditorUtility.OpenFilePanel(
                        "Choose Game Moment File",
                        "Moments/",
                        "json");
            if (path.Length != 0) {
                LoadGameFromDisc(path);
            }
        }
    }
}