using UnityEngine;
using System;
using System.Collections;
using UnityEditor;

namespace AT {

    /** Provides a custom inspector for the GameManager. 
     *
     * This is only used by developers. It should provide a simple and functional way to shift time (game saves), handle
     * game settings and work with integration tests.
     */
    [CustomEditor(typeof(GameManager))]
    public class GameManagerEditor : Editor {

        /** Draw custom Inspector interface.
         *
         * Unity GUI code is a state machine, so GUI.enabled controls all of the following GUI elements.
         * We use this to turn controls on/off depending on whether or not they are actually usable (i.e. you cannot
         * pause the game when it is already paused).
         */
        public override void OnInspectorGUI(){
            GameManager manager = (GameManager) target;
            DrawDefaultInspector(); // Draw default controls for anything that doesn't need to be customized
            DrawDebuggingAids(manager);
        }

        public static void DrawDebuggingAids(GameManager manager){
            if(manager == null){GUILayout.Label("Hit play to use GameManager debugging aids."); return;} // Manager isn't initialized yet?
            if(manager.disableDebugAids){GUILayout.Label("GameManager debugging aids turned off."); return;} 

            GUILayout.Label("Time Shifting", EditorStyles.boldLabel);
            GUILayout.Label("Save Interval (seconds): " + manager.SaveInterval.ToString());
            manager.SaveInterval = GUILayout.HorizontalSlider(manager.SaveInterval, manager.minSaveInterval, manager.maxSaveInterval);
            if(manager.saves.Count > 0){
                GUILayout.Label("Current save: " + (manager.currSave+1).ToString() + " / " + manager.saves.Count.ToString());
            }else{
                GUILayout.Label("Current save: -/-");
            }

            
            GUILayout.BeginHorizontal();

            if(manager.TimePaused()){ GUI.enabled = false;}
            if(GUILayout.Button("Pause")){ manager.PauseTime();}
            GUI.enabled = true;

            if(manager.TimePlaying()){ GUI.enabled = false;}
            if(GUILayout.Button("Resume")){ manager.ResumeTime();}
            GUI.enabled = true;

            if(!(manager.PrevSaveAvailable() && Application.isPlaying)){ GUI.enabled = false; } 
            if(GUILayout.Button("Previous")){ manager.LoadPrevSave();}
            GUI.enabled = true;

            if(!(manager.NextSaveAvailable() && Application.isPlaying)){ GUI.enabled = false; }
            if(GUILayout.Button("Next")){ manager.LoadNextSave();}
            GUI.enabled = true;

            GUILayout.EndHorizontal();


            GUILayout.BeginHorizontal();
             GUILayout.Label("Game Playthrough", EditorStyles.boldLabel);
            if(manager.saves.Count < 1){ GUI.enabled = false;}
            if(GUILayout.Button("Save")){
                manager.SavePlaythroughFile();
            }
            GUI.enabled = true;

            if(!Application.isPlaying){ GUI.enabled = false; }
            if(GUILayout.Button("Load")){
                manager.LoadPlaythroughFile();
            }
            GUI.enabled = true; 
            GUILayout.EndHorizontal();


            GUILayout.BeginHorizontal();
            GUILayout.Label("Game Moment", EditorStyles.boldLabel);
            if(!Application.isPlaying){ GUI.enabled = false; }
            if(GUILayout.Button("Save")){
                manager.SaveGameMomentFile();
            }
            if(GUILayout.Button("Load")){
                manager.LoadGameMomentFile();
            }
            GUI.enabled = true;
            GUILayout.EndHorizontal();

            GUILayout.BeginHorizontal();
            GUILayout.Label("Testing", EditorStyles.boldLabel);
            if(!Application.isPlaying){ GUI.enabled = false; }
            if(GUILayout.Button("Run Unit Tests")){
                manager.RunUnitTests();
            }
            if(GUILayout.Button("Run Integration Tests")){
                manager.RunIntegrationTests();
            }
            GUI.enabled = true;
            GUILayout.EndHorizontal();


            GUI.enabled = true; 
        }

    }
}