﻿using UnityEngine;
using UnityEditor;
using System;
using System.Collections;
using System.Collections.Generic;

namespace AT{
    namespace UBF{ 
        
        /** State of the world from an individual's point of view.
         */
        public class State{ 

            private GameObject agent;
            private AT.Vision vision;

            // THIS tank's info
            public TankState self;
            public Tuple<Vector2, Vector2> bodyWidth;
            public Tuple<Vector2, Vector2> bodyLength;
            public AT.Message[] messages; // messages received by this tank

            public List<TankState> visibleEnemyTanks;
            // first vector is position, second is velocity
            public List<Tuple<Vector2, Vector2>> visibleBullets; 
            // ALL walls are always known/visible
            public List<Tuple<Vector2, Vector2>> walls; 

            // GUI related toggles
            public bool showTankDimensionsFoldout;
            public bool showWallsFoldout;
            public bool showVisibleEnemyTanksFoldout;
            public bool showVisibleBulletsFoldout;
            public bool showSelfStateFoldout;
            public bool showMessagesFoldout;

            /** Use this for initialization
             */
            public State(UnityEngine.GameObject agent, AT.Vision vision){
                this.agent = agent;
                this.vision = vision;
                
                // Hide these by default due to their large size
                showTankDimensionsFoldout = false;
                showWallsFoldout = false; 
                // Show these by default to to their importance
                showVisibleEnemyTanksFoldout = true; 
                showVisibleBulletsFoldout = true;
                showSelfStateFoldout = true;
                showMessagesFoldout = true;
            }

            /**  Called by UBF.Controller once per tick. This can be used to get and set local state attributes if desired.
             */
            public void Update() {
                self = new TankState(agent);
                getTankDimensions();  
                messages = agent.GetComponent<AT.Tank>().getAndClearMessagesReceived().ToArray();

                visibleEnemyTanks = findAllVisibleEnemyTanks();
                visibleBullets = findAllVisibleBullets();
                walls = AT.MapManager.instance.Walls;                   
            }

            /** Determine dimension points (world reference) of the tank - width & length
             *
             * TODO: Not sure I understand this. Ask Stan Vernier about it (since he wrote it).
             */
            private void getTankDimensions(){
                BoxCollider2D collider = (BoxCollider2D) agent.gameObject.GetComponent<Collider2D>();

                Vector2 width1 = new Vector2(collider.offset.x ,collider.offset.y + (collider.size.y/2f));
                Vector2 width2 = new Vector2( collider.offset.x ,collider.offset.y - (collider.size.y / 2f));
                Vector2 length1 = new Vector2(collider.offset.x + collider.size.x / 2f, collider.offset.y);
                Vector2 length2 = new Vector2(collider.offset.x - collider.size.x / 2f, collider.offset.y);

                Vector2 worldWidth1 = agent.transform.TransformPoint(width1);
                Vector2 worldWidth2 = agent.transform.TransformPoint(width2);
                Vector2 worldLength1 = agent.transform.TransformPoint(length1);
                Vector2 worldLength2 = agent.transform.TransformPoint(length2);

                bodyWidth = new Tuple<Vector2, Vector2>(worldWidth1, worldWidth2);
                bodyLength = new Tuple<Vector2, Vector2>(worldLength1, worldLength2);
            }

            /** Find all visible enemy tanks.
             */
            private List<TankState> findAllVisibleEnemyTanks(){
                GameObject[] allTanks = GameObject.FindGameObjectsWithTag("Tank");
                List<TankState> visibleEnemyTanks = new List<TankState>();
                foreach(GameObject go in allTanks){
                    if(go == agent){
                        continue; // exclude ourselves
                    }else if(vision.isInConeOfVision(go.transform, 
                        LayerMask.GetMask(AT.GameManager.instance.TankLayer, AT.GameManager.instance.WallLayer))){
                        visibleEnemyTanks.Add(new TankState(go)); // is in our FOV
                    }
                }
                return visibleEnemyTanks;
            }

            /** Find all visible bullets and returns List of Tuples (Position, Velocity).
             */
            private List<Tuple<Vector2, Vector2>> findAllVisibleBullets(){
                GameObject[] allBullets = GameObject.FindGameObjectsWithTag("Bullet");
                List<Tuple<Vector2, Vector2>> visibleBullets = new List<Tuple<Vector2, Vector2>>(); // Tuple <Position, Velocity>
                foreach(GameObject go in allBullets){
                    if(vision.isInConeOfVision(go.transform, LayerMask.GetMask(AT.GameManager.instance.BulletLayer))){
                        // is in our FOV
                        Tuple<Vector2, Vector2> bullet = new Tuple<Vector2, Vector2>(go.transform.position, go.GetComponent<Rigidbody2D>().velocity);
                        visibleBullets.Add(bullet); 
                    }
                }
                return visibleBullets;
            }

            /** Draw a Custom view of this State object.
             *
             * Per Dr. Peterson's request, this allows the ControllerEditor to draw the current State in the Inspector.
             */
            public void DrawInspectorGUI(){ 
                GUILayout.BeginVertical();

                showSelfStateFoldout = EditorGUILayout.Foldout(showSelfStateFoldout, "This Tank's State");
                if(showSelfStateFoldout){                
                    self.DrawInspectorGUI();
                }

                showTankDimensionsFoldout = EditorGUILayout.Foldout(showTankDimensionsFoldout, "This Tanks' Dimensions");
                if(showTankDimensionsFoldout){ 
                    GUILayout.BeginVertical();
                    EditorGUILayout.LabelField("Body Width", 
                        String.Format("{0,6:F2}, {1,6:F2}    {2,6:F2}, {3,6:F2}", bodyWidth.first.x, bodyWidth.first.y, bodyWidth.second.x, bodyWidth.second.y), 
                        EditorStyles.boldLabel);

                    EditorGUILayout.LabelField("Body Length", 
                        String.Format("{0,6:F2}, {1,6:F2}    {2,6:F2}, {3,6:F2}", bodyLength.first.x, bodyLength.first.y, bodyLength.second.x, bodyLength.second.y), 
                        EditorStyles.boldLabel);
                    GUILayout.EndVertical();
                    GUILayout.EndHorizontal();
                }


                showVisibleEnemyTanksFoldout = EditorGUILayout.Foldout(showVisibleEnemyTanksFoldout, "Visible Enemy Tanks");
                if(showVisibleEnemyTanksFoldout){ 
                    for(int i=0; i<visibleEnemyTanks.Count; i++){
                        GUILayout.BeginHorizontal();
                        GUILayout.Space(30); // indent
                        GUILayout.BeginVertical();
                        visibleEnemyTanks[i].DrawInspectorGUI();
                        GUILayout.EndVertical();
                        GUILayout.EndHorizontal();
                    }
                }

                showVisibleBulletsFoldout = EditorGUILayout.Foldout(showVisibleBulletsFoldout, "Visible Bullets");
                if(showVisibleBulletsFoldout){ 
                    GUILayout.BeginHorizontal();
                    GUILayout.Space(30); // indent
                    GUILayout.BeginVertical();
                    foreach(Tuple<Vector2, Vector2> bullet in visibleBullets){
                        EditorGUILayout.LabelField("Position", 
                            String.Format("X {0,6:F2}  Y {1,6:F2}", bullet.first.x, bullet.first.y), 
                            EditorStyles.boldLabel);

                        EditorGUILayout.LabelField("Velocity", 
                            String.Format("X {0,6:F2}  Y {1,6:F2}", bullet.second.x, bullet.second.y), 
                            EditorStyles.boldLabel);
                    }
                    GUILayout.EndVertical();
                    GUILayout.EndHorizontal();
                }

                showWallsFoldout = EditorGUILayout.Foldout(showWallsFoldout, "Walls"); 
                if(showWallsFoldout){
                    GUILayout.BeginHorizontal();
                    GUILayout.Space(30); // indent
                    GUILayout.BeginVertical();
                    foreach(Tuple<Vector2, Vector2> w in walls){
                        GUILayout.Label(
                            String.Format("{0,7:F2}, {1,7:F2}    {2,7:F2}, {3,7:F2}", w.first.x, w.first.y, w.second.x, w.second.y),
                            EditorStyles.boldLabel);
                    }
                    GUILayout.EndVertical();
                    GUILayout.EndHorizontal();
                }

                showMessagesFoldout = EditorGUILayout.Foldout(showMessagesFoldout, "Messages Received");
                if(showMessagesFoldout){
                    GUILayout.BeginHorizontal();
                    GUILayout.Space(30); // indent
                    GUILayout.BeginVertical();
                    foreach(AT.Message m in messages){
                        GUILayout.Label(
                            m.getSender() + " (" + m.getSourceID() + ")  --> " + m.getReceiver(),
                            EditorStyles.boldLabel);
                        GUILayout.Label(m.getMessage());
                    }
                    GUILayout.EndVertical();
                    GUILayout.EndHorizontal();
                }

                GUILayout.EndVertical();
            }

        }
    }
}
