﻿using UnityEngine;
using UnityEditor;
using System.Collections;
using System.Collections.Generic;
using System.IO;
using System;

namespace AT{

	/** Represents current Map for AutoTank, including Walls.
	 *
	 *  Handles loading maps from file, clearing, etc.
	 *  
	 *  Singleton because there should only ever be one canonical instance.
	 */
	public class MapManager : MonoBehaviour {

		public static MapManager instance; // Static instance allows it to be accessed by any script

		public Material wallMaterial; 

		[HideInInspector]
		[SerializeField] 
		private string currMapFilePath;
		public string CurrMapFilePath{ // Note this property loads the map whenever the file path value changes
			get{return currMapFilePath;}
			set{
				// only take the filename (no directories) because we ASSUME it's in the Map directory
				// This forces the file paths to be portable between developers (instead of C:/Joe/AutoTank/Foo)
				currMapFilePath = "Maps/" + System.IO.Path.GetFileName(value); 
				LoadFromFile();
			}
		}

		[HideInInspector]
		[SerializeField]
		private string wallTag = "Wall";
		public string WallTag{
			get{return wallTag;}
			private set{wallTag = value;} // Private!
		}

		
		[HideInInspector]
		[SerializeField]
		private List<Tuple<Vector2, Vector2>> walls;
		public List<Tuple<Vector2, Vector2>> Walls{
			get{return walls;}
			private set{walls = value;} // Private !
		}

		/** Setup Walls list by finding the existing walls in the scene.
		 *
		 * Awake is, among other things, in charge of maintaining the Singleton. Note that if your
		 * code uses this Singleton, it may not exist during 'Awake()'. If you need to perform one-
		 * time setup, try using Start() instead.
		 *
		 * Adapted from https://unity3d.com/learn/tutorials/projects/2d-roguelike-tutorial/writing-game-manager
		 */
        void Awake(){ 
            if (instance == null){ // Check if instance already exists. If not, set instance to this
            	Debug.Log("MapManager starting...");
                instance = this;
            }else if (instance != this){ // If instance already exists and it's not this:
                //Then destroy this. This enforces our singleton pattern, meaning there can only ever be one instance of a MapManager.
                Destroy(gameObject); 
            }   
            
            // Setup the Walls List 
            FindExistingWallsInScene();
        }

		/** Removes all walls from the current scene.
		 *
		 * Doesn't rely on walls List to avoid missing walls that are manually inserted, etc.
		 */
		public void ClearMap(){
			GameObject[] oldWalls  = GameObject.FindGameObjectsWithTag(wallTag);
	        for(int i=0; i < oldWalls.Length; i++){
	            UnityEngine.Object.DestroyImmediate(oldWalls[i]); 
	        }
	        walls = new List<Tuple<Vector2, Vector2>>();
		}

		/** Reloads current map from file.
		 *
		 * Useful for when you're creating or changing a map file.
		 */
		public void ReloadMap(){
			LoadFromFile();
		}

		/** Creates a Vector-style Wall using a new GameObject, LineRenderer and EdgeCollider2D.
		 *
		 * This should be more efficient than dozens of sprite-based walls for our uses.
		 *
		 * Doesn't use a Rigidbody2D because it should not move at all.
		 */
		private void CreateWall(Vector2 start, Vector2 end, float width=0.1f, int zCoord=-1){
			GameObject wall = new GameObject(WallTag);
			wall.transform.parent = gameObject.transform; // set the MapManager GameObject as parent
			wall.tag = wallTag; // necessary for collision detection logic
			//wall.layer = LayerMask.GetMask(GameManager.instance.WallLayer);
			wall.layer = 10; // FIXME: Hardcoded layer number because the above isn't working

			LineRenderer lr = wall.AddComponent<LineRenderer>();
			lr.material = wallMaterial;
			Color c1 = Color.white;
	        lr.SetColors(c1, c1);
	        lr.SetWidth(width, width);
	        lr.SetVertexCount(2);
	        lr.useWorldSpace = true;
	        lr.SetPosition(0, new Vector3(start.x, start.y, zCoord));
	        lr.SetPosition(1, new Vector3(end.x, end.y, zCoord));
	        lr.receiveShadows = false;
	        lr.shadowCastingMode = UnityEngine.Rendering.ShadowCastingMode.Off;
	        lr.useLightProbes = false;

	        EdgeCollider2D ec = wall.AddComponent<EdgeCollider2D>();
	        ec.points = new Vector2[]{start, end};

	        Wall wallScript = wall.AddComponent<AT.Wall>();
	        wallScript.SetPoints(start, end);
		}

		/** Loads map from a text file in CSV format.
		 * 
		 * Example:
		 *     W,0,1,2,3   // would be a Wall from (0,1) to (2,3)
		 *
		 * Each wall is created using createWall
		 *
		 * If there is an error, it is logged to console, the map is cleared and the functions returns.
		 *
		 * This could be refactored into smaller functions, but I don't think it'll improve readability that much.
		 * Really, this encompasses all of the file's formatting, so I feel it's important to keep it together.
		 * Maybe I'll change my mind later (or when it gets more complex).
		 */
		private void LoadFromFile(){
			if(CurrMapFilePath == null){
            	Debug.LogError("Failed to load map. No map file path set.");
            	return;
            }
			Debug.Log("Loading map file: " + CurrMapFilePath);
			ClearMap(); // clear old Map first
			string[] lines;
			try{
				lines = System.IO.File.ReadAllLines(CurrMapFilePath);
			}catch(Exception){ // this is a bit too broad
				Debug.LogError("Failed to read file: " + CurrMapFilePath + " Clearing partially loaded map.");
				ClearMap(); 
				return;
			}
			string[] vals;
			string[] separators = {","};

			foreach(string line in lines){
				if(line.Substring(0,1) == "#"){
					continue; // This line is a comment, ignore it
				}

				vals = line.Split(separators, StringSplitOptions.None); 
				if(vals.Length < 1){
					fileLoadError("Each line of the map file must have at least a code (W or S). Invalid map file. Stopping.", line);
					return;
				}
				
				if(vals[0].Equals("W", StringComparison.OrdinalIgnoreCase)){
					if(vals.Length != 5){
						fileLoadError("Line coded as Wall was not a comma-separated five-tuple. Invalid map file. Stopping.", line);
						return;
					}
					try{
						Tuple<Vector2, Vector2> wall = new Tuple<Vector2, Vector2>(
									new Vector2(Convert.ToInt32(vals[1]), Convert.ToInt32(vals[2])), 
	   								new Vector2(Convert.ToInt32(vals[3]), Convert.ToInt32(vals[4]))
   						);
						walls.Add(wall);
						CreateWall(wall.first, wall.second);
					}catch(FormatException){
						fileLoadError("Wall 4-tuple failed conversion to Int32's. Invalid map file. Stopping.", line);
						return;
					}

				}else{
					fileLoadError("Invalid line code. Invalid map file. Stopping.", line);
					return;
				}
			}
		}

		/** Log the Error reason, line (input) and clear the map.
		 *
		 *  I refactored this out of the LoadMapFile function. Not sure if it's really any better.
		 */
		private void fileLoadError(string reason, string line){
			Debug.LogError(reason);
			Debug.Log("Line: " + line);
			ClearMap();
		}

		/** Find existing Walls in the current Scene and set the Walls list to match.
		 */
		private void FindExistingWallsInScene(){
			// Setup new list to hold their positions
			walls = new List<Tuple<Vector2, Vector2>>();

			// Find and add each Wall to the list using their Wall script
			GameObject[] oldWalls  = GameObject.FindGameObjectsWithTag(wallTag);
	        for(int i=0; i < oldWalls.Length; i++){
	           Wall wallScript = oldWalls[i].GetComponent<Wall>() as Wall; 
	           walls.Add(wallScript.GetVertices());
	        }   
		}
	}
}
