﻿using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using System;

namespace AT {

    /** Base AutoTank Tank functionality.
     *
     *  Controls basic physics, health and vision, along with controls (DriveForward, Turn, etc).
     */
    public class Tank : MonoBehaviour {

        [HideInInspector]
        public static float trackRotationSpeed = 3.0f; // PlayerController needs this info which is why this is static 
        [HideInInspector]
        public static float turretRotationSpeed = 3.0f; // PlayerController needs this info which is why this is static 
        [HideInInspector]
        public static float maxMoveSpeed = 3.5f;
        [HideInInspector]
        public float reloadTimeInSeconds = 0.5f;

        public int hitPoints = 100;
        [HideInInspector]
        public int bulletDamage = 10;
        [HideInInspector]
        public int tankDamage = 20;

        [HideInInspector]
        public Transform turret_transform;   // Used by RotateTurret and Fire
        [HideInInspector]
        public AT.Vision vision;
        protected Transform turret_end_point_transform;   // Used by Fire
        [HideInInspector]
        public AT.Health health;

        private float lastShotTime = 0.0f;
        private bool oldCanFireGun = false; // see UpdateCanFireGun

        private bool loadedFromDict = false;

        public string tankColor = null; // unique color assigned to this Tank by GameManager can be used as ID
        public AT.Team team = AT.Team.None; // if tanks are on teams, this defines which team they are on

        public List<AT.Message> messagesReceived;
        public static void SendMessage(AT.Message msg){
            GameObject[] tankGOs = GameObject.FindGameObjectsWithTag("Tank");
            for(int i=0; i < tankGOs.Length; i++){
                AT.Tank tank = tankGOs[i].GetComponent<AT.Tank>();

                if(msg.getReceiverType() == AT.Message.ReceiverType.Team){
                    if(tank.team.ToString() == msg.getReceiver()){
                        tank.messagesReceived.Add(msg);
                    }
                }else if(msg.getReceiverType() == AT.Message.ReceiverType.Individual){
                    if(tank.tankColor == msg.getReceiver()){
                        tank.messagesReceived.Add(msg);
                    }
                }else{
                    Debug.LogError("Message type is not one of the expected types! This code probably needs updated!");
                }
            }
        }
        /** Return the current List of Messages and then delete the original list for this tank.
         *
         *  This is meant to keep processed/old messages from piling up for the tank once State retrieves them. 
         */
        public List<AT.Message> getAndClearMessagesReceived(){
            List<AT.Message> messages = messagesReceived;
            messagesReceived = new List<AT.Message>();
            return messages;
        }


        /** Initial setup for Tank. 
         *
         * Finds and saves reference to several key child GameObjects for later use (Turret, 
         * TurretEndPoint, Health and Vision). This makes using them much faster.
         */
        public virtual void Awake(){
            messagesReceived = new List<AT.Message>();

            foreach(Transform t in transform){
                if(t.name == "Turret"){
                    turret_transform = t;
                    break;
                }
            }
            foreach(Transform t in turret_transform.transform){
                if(t.name == "TurretEndPoint"){
                    turret_end_point_transform = t;
                    break;
                }
            }

            health = gameObject.GetComponent<AT.Health>() as AT.Health;
            if(!loadedFromDict){
                health.config(hitPoints, hitPoints, bulletDamage, tankDamage);
            }
            vision = turret_transform.gameObject.GetComponent<AT.Vision>() as AT.Vision;
        }

        public virtual void UpdateState(){
            UpdateCanFireGun();
        }

        public int getHealth(){
            return health.getHealth();
        }

        /** Can we fire the gun now? (Reload Time)
         */
        public bool CanFireGun(){
            return (Time.time - lastShotTime > reloadTimeInSeconds);
        }

        /** Stores old state of CanFireGun so we only Broadcast changes when this changes.
         *
         * Sends updates via BroadcastMessage("UpdateCanFireGun", bool canFire) to those interested.
         */
        private void UpdateCanFireGun(){
            bool canFire = CanFireGun();
            if(canFire != oldCanFireGun){
                oldCanFireGun = canFire;
                BroadcastMessage("UpdateCanFireGun", canFire); 
            }
        }

        /** Turn the Tank's Turret. Input is bound from -1.0 to 1.0
         *
         *  1.0 is full speed right (0.5 if 50% full speed right)
         *  -1.0 is full speed left (-0.5 if 50% full speed left)
         *  0.0 is a no-op (do not turn)
         *
         * DESIGN NOTE:
         * Rotating child elements (the turret) is by default relative to the parent (the tracks/tank body). That means 
         * the turret will rotate FASTER when rotating in the same direction as the tracks! This "corrects" this 
         * behavior by factoring in any track rotation, meaning the turret will turn at the same speed no matter if it 
         * is turning the opposite or same direction as the tracks. This isn't necessarily "correct" but it's how I've 
         * chosen to have the game behave. 
         */
        protected void RotateTurret(float turnRight, Vector3 tracksTurned){
            // Vector3.back is the Z-axis aimed AT the camera
            turnRight = Mathf.Clamp(turnRight, -1.0f, 1.0f);
            //turret_transform.Rotate(Vector3.back * turretRotationSpeed * turnRight); // WITHOUT track rot correction
            turret_transform.Rotate(Vector3.back * turretRotationSpeed * turnRight - tracksTurned); // WITH track rot correction
        }

        /** Move forward or backward. Input is bound from -1.0 to 1.0
         *
         *  1.0 is full speed forward (0.5 if 50% full speed forward)
         *  -1.0 is full speed backward (-0.5 if 50% full speed backward)
         *  0.0 is a no-op (do not move)
         */
        protected void DriveForward(float driveForward){
            driveForward = Mathf.Clamp(driveForward, -1.0f, 1.0f);
            // We use transform.right because everything assume right~0 degrees, so that's forward by default
            GetComponent<Rigidbody2D>().velocity = new Vector2(transform.right.x * maxMoveSpeed * driveForward, 
                                                               transform.right.y * maxMoveSpeed * driveForward);    
        }

        /** Turn the Tank's Tracks. Input is bound from -1.0 to 1.0
         *
         *  1.0 is full speed right (0.5 if 50% full speed right)
         *  -1.0 is full speed left (-0.5 if 50% full speed left)
         *  0.0 is a no-op (do not turn)
         */
        protected Vector3 TurnTracks(float turnRight){
            // Need to rotate around the Z axis ONLY (2D's Left and Right)
            turnRight = Mathf.Clamp(turnRight, -1.0f, 1.0f); // -1.0 is left and 1.0 is right
            Vector3 rotateAmt = Vector3.back * trackRotationSpeed * turnRight; // Vector3.back is the Z-axis aimed AT the camera
            transform.Rotate(rotateAmt);
            return rotateAmt;
        }

        /** Fire the gun if enough time has passed since the last firing (reloadTimeInSeconds).
         *
         * Uses the bulletPrefab which must be set in the Unity Inspector on the GameManger.
         */
        protected void FireGun(){
            if(CanFireGun()){
                lastShotTime = Time.time;
                // The Bullet script handles its movement and logic
                Instantiate(GameManager.instance.bulletPrefab, turret_end_point_transform.position, turret_transform.rotation);
            }
        }

        /** Method to set the GameObject's name attribute which is used in the Unity Hierarchy view.
         *
         * Appends the team name in parenthesis.
         */
        protected void SetName(string name){
            this.name = name + " (Team " + team + ")";
        }

        /** Save this Tanks internal state required for persistence to a Dictionary.
         *
         *  This information MUST satisfy the LoadFromDict() method.
         */
        public virtual Dictionary<string, object> ToDict(){
                Dictionary<string, object> tank = new Dictionary<string, object>();

                tank["_type"] = this.GetType().FullName; // mostly just an FYI; overwritten by derived classes
                tank["health"] = health.getHealth();
                tank["lastShotTime"] = lastShotTime;
                tank["color"] = tankColor;
                tank["team"] = team; // Enum will be converted to string representation

                tank["posX"] = transform.position.x;
                tank["posY"] = transform.position.y;
                tank["posZ"] = transform.position.z;

                // Rotation uses Quaternions internally, so we need to store all values (X,Y,Z,W) to
                // prevent screwing it up despite only rotating around the Z axis
                tank["rotX"] = transform.rotation.x;
                tank["rotY"] = transform.rotation.y;
                tank["rotZ"] = transform.rotation.z;
                tank["rotW"] = transform.rotation.w;

                tank["turretRotX"] = turret_transform.rotation.x;
                tank["turretRotY"] = turret_transform.rotation.y;
                tank["turretRotZ"] = turret_transform.rotation.z;
                tank["turretRotW"] = turret_transform.rotation.w;

                return tank;
        }

        public virtual void FromDict(Dictionary<string, object> tank){
            loadedFromDict = true;
            health.config(((int) (long) tank["health"]), hitPoints, bulletDamage, tankDamage);
            lastShotTime = (float) ((double)tank["lastShotTime"]);
            tankColor = (string) tank["color"];
            team = (AT.Team) Enum.Parse( typeof(AT.Team), (string) tank["team"] ); 

            transform.position = new Vector3((float) ((double)tank["posX"]), (float) ((double)tank["posY"]), (float) ((double)tank["posZ"]));
            transform.rotation = new Quaternion((float) ((double)tank["rotX"]), (float) ((double)tank["rotY"]), (float) ((double)tank["rotZ"]), (float) ((double)tank["rotW"]));
            turret_transform.rotation = new Quaternion((float) ((double)tank["turretRotX"]), (float) ((double)tank["turretRotY"]), 
                                                       (float) ((double)tank["turretRotZ"]), (float) ((double)tank["turretRotW"]));
        }


        
    }
}
