﻿using UnityEngine;
using System.Collections;
using System;

namespace AT{
    /** Vision script for both AI and human controlled tanks.
     *
     * For AI, this is used to determine if an object is in sight.
     * For humans, this is used to draw a Field of View (FOV) cone. 
     * 
     * Note: Relies on Wall and it's vision blocking. In Update(), this calls DrawWVB(playerPosition) on each 
     *       wall using SendMessage.
     *       To speed this up, we store a reference to each wall object (see Awake()). 
     *
     * Adapted from Stack Exchange (The MIT License) - Copyright 2014 Rickz0r:
     *     http://gamedev.stackexchange.com/a/70158
     *     http://gamedev.stackexchange.com/questions/31170/drawing-a-dynamic-indicator-for-a-field-of-view
     */
    public class Vision : MonoBehaviour {
        
        private float angle_fov;
        private float dist_min;
        private float dist_max;
        private Material material;  // This is used for the Human FOV "fog of war" and is just BLACK

        private int quality = 60; // higher numbers use more triangles to draw the Vision mesh and look better
        private Mesh coneOfVisionMesh;
        private GameObject[] walls;
        private AT.UBF.Controller controller;
        
        /** Checks that the Transform is within our Field of Vision and not blocked by a wall or tank.
         *
         * LayerMask includeLayers: Only detect Transforms on these Layers (ignore all others)
         *     Example: use   LayerMask.GetMask(GameManager.instance.TankLayer, GameManager.instance.WallLayer)   for Tanks and Walls
         *               of   LayerMask.GetMask(GameManager.instance.BulletLayer)   for Bullets
         * 
         * 
         * Note: Requires "Raycasts Start In Colliders" to be unchecked (false)
         *       Go to Edit > Project Settings > Physics 2D
         *
         * BUGFIX: This uses a raycast using ONLY the colliders on the default layer, which excludes Bullets (which are 
         *       now on the BulletLayer). This was causing the AI to "lose sight" of the enemy after firing at them!
         */
        public bool isInConeOfVision(Transform obj, LayerMask includeLayers){
            Vector3 rayDirection = obj.position - transform.position;
            RaycastHit2D hit = Physics2D.Raycast(transform.position, // Origin
                                         rayDirection,
                                         VisionManager.instance.playerVisionDistMax, // Max Distance of Ray
                                         includeLayers); // Filter 

            if(hit.transform == obj){ // if not obstructed by walls or other tanks ?
                if(Vector3.Distance(transform.position, obj.position) < dist_min){ // if it is very close
                    if(VisionManager.instance.drawVisionHits){
                        Debug.DrawLine(transform.position, obj.position, Color.red);
                    }
                    return true; 
                }
                float angle = Vector3.Angle(rayDirection, transform.right);

                if(angle  < angle_fov/2){ // within our Field of View
                    if(VisionManager.instance.drawVisionHits){
                        Debug.DrawLine(transform.position, obj.position, Color.red);
                    }
                    return true; 
                }
            }
            return false;
        }

        /** Initial setup. 
         *
         * Pulls in VisionManager settings only once and sets up player controlled Field of View 
         * mesh. The mesh is then updated every tick as needed. This is supposed to be faster than
         * recreating it each time. Note that for AI tanks, the mesh update will NEVER happen.
         */
        public void Start(){
            try{ // Pull the global vision settings from the manager upon awake
                angle_fov = VisionManager.instance.playerVisionAngleFOV;
                dist_min = VisionManager.instance.playerVisionDistMin;
                dist_max = VisionManager.instance.playerVisionDistMax;
                material = VisionManager.instance.visionBlockingMaterial;
                // pull the parent GameObject's Controller script so we can reference the behavior 
                // (for checking whether or not this is PlayerController or not; need to draw FOV)
                controller = (AT.UBF.Controller) gameObject.GetComponentInParent<AT.UBF.Controller>();
            }catch(NullReferenceException){
                Debug.LogError("VisionManager not found. Make sure you have one VisionManager component in the scene.");
                return;
            }

            // 1. Find all wall objects so we can later update their vision blocking
            walls = GameObject.FindGameObjectsWithTag( MapManager.instance.WallTag );

            // 2. Setup Mesh, normals, and uv which only needs done once
            coneOfVisionMesh = new Mesh();
            coneOfVisionMesh.vertices = new Vector3[4 * quality];   

            Vector2[] uv = new Vector2[4 * quality];
            for (int i = 0; i < uv.Length; i++)
                uv[i] = new Vector2(0, 0);
            coneOfVisionMesh.uv = uv;

            Vector3[] normals = new Vector3[4 * quality];
            for (int i = 0; i < normals.Length; i++)
                normals[i] = new Vector3(0,1,0);
            coneOfVisionMesh.normals = normals;

            // Set triangles, which is simply the order of the vertices and only needs done once
            int[] triangles = new int[3 * 2 * quality];
            for(int i = 0; i < quality; i++){
                int a = 4 * i;
                int b = 4 * i + 1;
                int c = 4 * i + 2;
                int d = 4 * i + 3;

                triangles[6 * i] = a;       // Triangle 1: abc
                triangles[6 * i + 1] = b;
                triangles[6 * i + 2] = c;
                triangles[6 * i + 3] = c;   // Triangle 2: cda
                triangles[6 * i + 4] = d;
                triangles[6 * i + 5] = a;
            }
            coneOfVisionMesh.triangles = triangles;
        }       

        /** Update the Field of View (FOV) if this is a Player Controlled Tank and the DrawFOV 
         * option is on. Otherwise, do nothing.
         */
        void Update(){
            if(controller == null || controller.behavior == null){Util.QuitImmediately();}
            // Allows us to turn this on/off while debugging by getting it EVERY update (but expensive)
            // This is not clean at all; we're referencing bits that shouldn't concern us and using a "Magic String"
            if(controller.behavior.GetType().FullName != "AT.UBF.Behaviors.PlayerControlled" ||
               !VisionManager.instance.drawPlayerControlledFOV){ 
                return; // AI don't need a visible cone of vision; they use isInConeOfVision()
            }

            // Update walls vision blocking first (with player's current position)
            for(int i=0; i < walls.Length; i++){
                walls[i].SendMessage("DrawVB", transform.position);
            }

            float angle_lookat = transform.eulerAngles.z;

            float angle_start = angle_lookat + angle_fov/2;
            //float angle_end = angle_lookat - angle_fov/2; // FIXME: not used?
            float angle_delta = (360.0f-angle_fov) / quality;

            float angle_curr = angle_start;
            float angle_next = angle_start + angle_delta;

            Vector3 pos_curr_min = Vector3.zero;
            Vector3 pos_curr_max = Vector3.zero;

            Vector3 pos_next_min = Vector3.zero;
            Vector3 pos_next_max = Vector3.zero;

            // Unity documentation recommends getting old vertices and changing them for efficiency
            // http://docs.unity3d.com/ScriptReference/Mesh.html
            Vector3[] vertices = coneOfVisionMesh.vertices; 

            for (int i = 0; i < quality; i++){
                Vector3 sphere_curr = new Vector3(
                    Mathf.Cos(Mathf.Deg2Rad * (angle_curr)),
                    Mathf.Sin(Mathf.Deg2Rad * (angle_curr)),
                    0);

                Vector3 sphere_next = new Vector3(
                    Mathf.Cos(Mathf.Deg2Rad * (angle_next)),
                    Mathf.Sin(Mathf.Deg2Rad * (angle_next)),
                    0);

                pos_curr_min = transform.position + sphere_curr * dist_min;
                pos_curr_max = transform.position + sphere_curr * dist_max;

                pos_next_min = transform.position + sphere_next * dist_min;
                pos_next_max = transform.position + sphere_next * dist_max;

                int a = 4 * i;
                int b = 4 * i + 1;
                int c = 4 * i + 2;
                int d = 4 * i + 3;

                vertices[a] = pos_curr_min;
                vertices[b] = pos_curr_max;
                vertices[c] = pos_next_max;
                vertices[d] = pos_next_min;

                angle_curr += angle_delta;
                angle_next += angle_delta;
            }
            coneOfVisionMesh.vertices = vertices;

            Graphics.DrawMesh(coneOfVisionMesh, Vector3.zero, Quaternion.identity, material, 0);
        }

    }
}
