using UnityEngine;
using System.Collections;
using UnityEditor;

namespace AT {

    /** Singleton to manage the game's Vision blocking (Fog of War).
     *
     * This holds global vision settings. This is easier than changing each prefab's Vision script
     * each time we change something. Note that these are typically only pulled in at the START of 
     * the game. So changing them during play will not change anything. DrawPlayerControlledFOV is
     * the only exception.
     *
     */
    public class VisionManager : MonoBehaviour {

        public static VisionManager instance; // Static instance allows it to be accessed by any script

        public float playerVisionAngleFOV = 40;
        public float playerVisionDistMin  = 1.0f;
        public float playerVisionDistMax  = 50.0f;  

        [Header("Prefabs and Materials")]
        public Material visionBlockingMaterial; // Simple black "fog" Material 

        [Header("Player-Controlled Only Settings")]
        public bool  drawPlayerControlledFOV  = true;
        public int quality = 60; // higher numbers use more triangles to draw the Vision mesh 
        public float wallDistMax = 50f;

        // Draws line when isInConeOfVision returns true. Will only show on Scene view (not Game view)
        [HideInInspector]
        public bool drawVisionHits = false; 

        /** 
         *
         * Awake is, among other things, in charge of maintaining the Singleton. Note that if your
         * code uses this Singleton, it may not exist during 'Awake()'. If you need to perform one-
         * time setup, try using Start() instead.
         *
         * Adapted from https://unity3d.com/learn/tutorials/projects/2d-roguelike-tutorial/writing-game-manager
         */
        void Awake(){ 
            if (instance == null){ // Check if instance already exists. If not, set instance to this
                Debug.Log("VisionManager starting...");
                instance = this;
            }else if (instance != this){ // If instance already exists and it's not this:
                // Then destroy this. This enforces our singleton pattern, meaning there can only 
                // ever be one instance of a VisionManager.
                Destroy(gameObject); 
            }   
        }
    }
}
