using UnityEngine;
using System.Collections;
using System;

namespace AT{

    /** Stores the end points and handles vision blocking resulting from this wall for Player-Controlled agents. 
     */
    public class Wall : MonoBehaviour {
        // these should be the same across all objects
        private float distMax = 50f;
        private Material material;  

        private Mesh mesh;
        public Vector3 vertexA;
        public Vector3 vertexB;

        public void SetPoints(Vector3 vertexA, Vector3 vertexB){
            this.vertexA = vertexA;
            this.vertexB = vertexB;
        }

        /** Grab settings from VisionManager and setup initial mesh for faster updates in DrawVB.
         *
         * VisionManager might not exist yet in Awake, so this MUST be in Start (due to uncertain call order).
         */
        void Start(){
            // Pull the global vision settings from the manager
            try{ 
                material = VisionManager.instance.visionBlockingMaterial;
                distMax = VisionManager.instance.wallDistMax;
            }catch(NullReferenceException){
                Debug.LogError("VisionManager not found. Make sure you have one VisionManager component in the scene.");
                return;
            }

            // Setup Mesh for faster updates
            mesh = new Mesh();
            mesh.vertices = new Vector3[4]; 
    
            Vector2[] uv = new Vector2[4];
            for (int i = 0; i < uv.Length; i++)
                uv[i] = new Vector2(0, 0);
            mesh.uv = uv;

            Vector3[] normals = new Vector3[4];
            for (int i = 0; i < normals.Length; i++)
                normals[i] = new Vector3(0,1,0);
            mesh.normals = normals;

            // Set triangles, which is simply the order of the vertices and only needs done once
            int[] triangles = new int[6];
            triangles[0] = 0;   // Triangle 1: abc
            triangles[1] = 1;
            triangles[2] = 2;
            triangles[3] = 2;   // Triangle 2: cda
            triangles[4] = 3;
            triangles[5] = 0;
            mesh.triangles = triangles;
        }       

        /** Return a simplified representation of this Wall's vertices.
         */
        public Tuple<Vector2, Vector2> GetVertices(){
            return new Tuple<Vector2, Vector2>(vertexA, vertexB);
        }

        /** Draws a vision blocking mesh (for everything behind this wall) based on the Player's position.
         *
         * This is a bit of a hack, but if a Player-Controlled agent exists, this creates a Fog of War mesh which blocks 
         * the player's vision for anything behind this wall. The Player-Controlled agent sends a "DrawVB" message to all
         * Walls with its current position for this to work. This is more efficiently than polling for PlayerControlled
         * agents.
         */
        void DrawVB(Vector3 playerPosition){
            // Unity documentation recommends getting old vertices and changing them for efficiency
            // http://docs.unity3d.com/ScriptReference/Mesh.html
            Vector3[] vertices = mesh.vertices; 

            vertices[0] = vertexA;
            vertices[1] = vertexB;
            vertices[2] = (vertexB - playerPosition) * distMax; 
            vertices[3] = (vertexA - playerPosition) * distMax;

            mesh.vertices = vertices;

            Graphics.DrawMesh(mesh, Vector3.zero, Quaternion.identity, material, 0);
        }
    }
}
