﻿using UnityEngine;
using System;
using System.Collections;
using System.Collections.Generic;


/** Utility for checking for two lines intersecting and determining the angle of the two 
 * lines if an intersection is found
 *
 * Written by Stan Vernier
 */
class LineIntersection {

    /* If there's an intersection, calculates the shortest turning angle to avoid line2.  
     * If no intersection return null.
     */
    public static float? calculate(Tuple<Vector2, Vector2> line1, Tuple<Vector2, Vector2> line2) {
        if(doLinesIntersect(line1, line2)){
            return calculateAngle(line1, line2);
        }
        return null;
    }

    public static float calculateAngle(Tuple<Vector2, Vector2> line1, Tuple<Vector2, Vector2> line2){
        float angle1 = Util.getRelativeAngleToTarget(line1.first, line2.first);
        float angle2 = Util.getRelativeAngleToTarget(line1.first, line2.second);

        float left, right;
        if (angle1 < 180f) {
            right = angle1;
            left = angle2;
        } else {
            right = angle2;
            left = angle1;
        }
        //want negative value if to the left
        left -= 360;
        if (right < Mathf.Abs(left)) {
            return right;
        } else {
            return left;
        }
    }

    public static float orientation(Vector2 a, Vector2 b, Vector2 c) {
        float val = (b.y - a.y) * (c.x - b.x) - (b.x - a.x) * (c.y - b.y);
        if (val == 0) return 0;

        return (val > 0) ? 1 : 2;
    }

    public static bool onSegment(Vector2 a, Vector2 b, Vector2 c){
        if(b.x <= Math.Max(a.x, c.x) && b.x >= Math.Min(a.x, c.x) &&
            b.y <= Math.Max(a.y, c.y) && b.y >= Math.Min(a.y, c.y)){
            return true;
        }
        return false;
    }

    public static bool doLinesIntersect(Tuple<Vector2, Vector2> line1, Tuple<Vector2, Vector2> line2) {
        float o1 = orientation(line1.first, line1.second, line2.first);
        float o2 = orientation(line1.first, line1.second, line2.second);
        float o3 = orientation(line2.first, line2.second, line1.first);
        float o4 = orientation(line2.first, line2.second, line1.second);

        if (o1 != o2 && o3 != o4) {
            return true;
        }

        if (o1 == 0 && onSegment(line1.first, line2.first, line1.second)) {
            return true;
        }
        if (o2 == 0 && onSegment(line1.first, line2.second, line1.second)) {
            return true;
        }
        if (o3 == 0 && onSegment(line2.first, line1.first, line2.second)) {
            return true;
        }
        if (o4 == 0 && onSegment(line2.first, line1.second, line2.second)) {
            return true;
        }
        return false;
    }
}
