using UnityEngine;
using UnityEditor;
using System.Collections;

/** Simple user input pop-up dialog box with a prompt and text area for the user's response.
 *
* After the user enters any text they wish (in response to the prompt) the provided callback will be sent that text.
 * i.e. callbackSendMessage(callbackName, userInput)
 */
public class UserInputDialog : EditorWindow{

    private string userPrompt;
    private string userInput;

    // These are used for the callback after the user has hit submit i.e. callback.SendMessage(callbackName, userInput)
    private MonoBehaviour callback;
    private string callbackName;

    private Vector2 scrollPos;
    private GUIStyle promptStyle;
    

    /** Create a new instance of this pop-up dialog box with the provided prompt.
     */
    public static UserInputDialog Init(string userPrompt, MonoBehaviour callback, string callbackName, int width=400, int height=200){
        UserInputDialog window = (UserInputDialog) ScriptableObject.CreateInstance<UserInputDialog>();
        window.position = new Rect(Screen.width / 2, Screen.height / 2, width, height);
        window.SetupPromptStyle();

        window.userPrompt = userPrompt;
        window.callback = callback;
        window.callbackName = callbackName;

        window.ShowPopup();
        return window;
    }

    void SetupPromptStyle(){
        promptStyle = new GUIStyle();
        promptStyle.wordWrap = true;
        promptStyle.fontStyle = FontStyle.Bold;
        promptStyle.margin = new RectOffset(5,5,5,10);

        EditorStyles.textField.wordWrap = true;
    }

    /** Draw the GUI, and send the user's input to the callback when they hit "Submit"
     */
    void OnGUI(){        
        EditorGUILayout.LabelField(userPrompt, promptStyle);

        scrollPos = EditorGUILayout.BeginScrollView(scrollPos);
        userInput = EditorGUILayout.TextArea(userInput, GUILayout.ExpandHeight(true));
        EditorGUILayout.EndScrollView();

        if(GUILayout.Button("Submit") && userInput.Length > 0){ 
            this.Close();
            callback.SendMessage(callbackName, userInput);
        }
    }
}