using UnityEngine;
using UnityEditor;
using System.Collections;
using System.Collections.Generic;
using System;
using System.IO;


/** Utility methods for Unity.
 */
public static class Util{

    /** Quit the game immediately. If quitting fails, use Debug.Break() to pause the game.
     *
     *  This does not save any data, or perform any other end game logic. This is typically only used when we've 
     *  encountered an unrecoverable error (i.e. failing to load an Agent's UBF Tree).
     */
    public static void QuitImmediately(){
        Debug.Log("QUIT IMMEDIATELY!");
        #if UNITY_EDITOR
            EditorApplication.ExecuteMenuItem("Edit/Play");
         #else
            Application.Quit();
         #endif

        if(UnityEditor.EditorApplication.isPlaying){
            // Hitting Play didn't work, so let's try breaking (will pause instead of quit)
            Debug.Break();
        }
    }

    public static void JsonDictToFile(string path, Dictionary<string, object> dict){
        string json = MiniJSON.Json.Serialize(dict);
        System.IO.File.WriteAllText(path, json);
    }
    
    public static Dictionary<string, object> JsonDictFromFile(string path){
        string json;
        try{
            json = System.IO.File.ReadAllText(path);
        }catch(FileNotFoundException){
            Debug.LogError("JSON file not found: " + path);
            return null;
        }
        try{
            return MiniJSON.Json.Deserialize(json) as Dictionary<string, object>;
        }catch(Exception){
            Debug.LogError("Failed to deserialize JSON string.");
            return null;
        };
    }
    
    /** Returns angle in degrees from reference Vector to target Vector.
     *
     * Example 1: 
     *      Get the angle from tank A to tank B
     *      float angle = getRelativeAngleToTarget(tankA.transform.position, tankB.transform.position);
     *
     * Example 2:
     *      Get the angle from our tank (given its' state object) to the mouse
     *      Vector2 mousePos = Camera.main.ScreenToWorldPoint(Input.mousePosition);
     *      float angle = getRelativeAngleToTarget(state.position, mousePos);
     */
    public static float getRelativeAngleToTarget(Vector2 reference, Vector2 target){
        float deltaY = target.y - reference.y;
        float deltaX = target.x - reference.x;
        float angleInDegrees = Mathf.Atan2(deltaY, deltaX) * Mathf.Rad2Deg;
        if(angleInDegrees < 0){angleInDegrees += 360f;}
        return angleInDegrees;
    }

    /** Find a turnRight value to rotate our turret to the target position (pure pursuit aka no lead or lag).
     *
     * Compare targetRotation to our turret rotation, return a turnRight vote that will make them match 
     * (if difference is less than a single full speed roation (maxRot), return what fraction of full speed 
     * we should move.
     */
    public static float findShortestTurretRotationToTarget(Vector2 myPos, Vector2 targetPos, float myTurretRotation){
        float targetRotation = Util.getRelativeAngleToTarget(myPos, targetPos);         
        float maxRot = AT.Tank.turretRotationSpeed;
        return findShortestRotation(targetRotation, myTurretRotation, maxRot);
    }

    /** Find a turnRight value to rotate our tracks to the target position (pure pursuit aka no lead or lag).
     *
     * Compare targetRotation to our track rotation, return a turnRight vote that will make them match 
     * (if difference is less than a single full speed roation (maxRot), return what fraction of full speed 
     * we should move.
     */
    public static float findShortestTrackRotationToTarget(Vector2 myPos, Vector2 targetPos, float myTrackRotation){
        float targetRotation = Util.getRelativeAngleToTarget(myPos, targetPos);             
        float maxRot = AT.Tank.trackRotationSpeed;
        return findShortestRotation(targetRotation, myTrackRotation, maxRot);
    } 

    /** Helper method to determine if we should turn right or left to match the target rotation as fast as possible.
     *
     * 
     *  maxRotation is used to determine if we should move full speed (-1.0 or 1.0) or slower depending on 
     *  whether we can reach the target rotation this tick.
     *
     *  This code was pulled out of findShortest(Track/Turret)RotationToTarget()
     */
    private static float findShortestRotation(float targetRotation, float currRotation, float maxRotation){
        float temp = (targetRotation - currRotation + 360) % 360;
        float turnRight = 0.0f;

        if (temp > 180) {
            if(temp > 360-maxRotation){ 
                turnRight = Mathf.Abs(360-temp)/maxRotation;
            }else{
                turnRight = 1.0f;
            }
        }else{
            if(temp < maxRotation){
                turnRight = (0-temp)/maxRotation;
            }else{
                turnRight = -1.0f;
            }
        }

        return turnRight;
    }

    /** Takes a screen Rectangle and centers it using the screen's width and height.
     */
    public static Rect centerRect(Rect r){
        r.x = ( Screen.width  - r.width ) / 2;
        r.y = ( Screen.height - r.height ) / 2;
        return r;
    }

    /** Returns the current Date and Time in a reliably sortable format (as a string).
     *
     * Note: This format replaces the colons (:) with underscores to allow these to be used for file paths.
     *
     * See this page for more: https://msdn.microsoft.com/en-us/library/az4se3k1(v=vs.110).aspx#Sortable
     */
    public static string getSortableDateTimeString(){
        DateTime localDate = DateTime.Now;
        return localDate.ToString("yyyy'-'MM'-'dd'T'HH'_'mm'_'ss");
    }

    /** Helper method for distToLineSegment.
     */
    public static float distSquared(Vector2 v, Vector2 w){
        return Mathf.Pow(v.x - w.x, 2) + Mathf.Pow(v.y - w.y, 2);
    }

    /** Return the distance from point p to the line segment (v, w) AND the closest point on that line segment.
     *
     * Code adapted from Stack Overflow (The MIT License) - Copyright 2010-2016 Grumdrig:
     *     http://stackoverflow.com/a/1501725
     */
    public static Tuple<float, Vector2> distToLineSegment(Vector2 point, Vector2 lineSegPointA, Vector2 lineSegPointB){
        // Return minimum distance between line segment vw and point p
        float lenSquared = distSquared(lineSegPointA, lineSegPointB); // i.e. |w-v|^2 -  avoid a sqrt
        if(lenSquared == 0.0){ // v == w case
            // The line segment has a length of zero, so the distance to it is the distance to either or its' points
            return new Tuple<float, Vector2>(Mathf.Sqrt(distSquared(point, lineSegPointA)), lineSegPointA);
        }
        // Consider the line extending the segment, parameterized as v + t (w - v).
        // We find projection of point p onto the line. 
        // It falls where t = [(p-v) . (w-v)] / |w-v|^2
        float t = ((point.x - lineSegPointA.x) * (lineSegPointB.x - lineSegPointA.x) + (point.y - lineSegPointA.y) * (lineSegPointB.y - lineSegPointA.y)) / lenSquared;
        if(t < 0.0){ // Beyond the 'v' end of the segment
            return new Tuple<float, Vector2>(Mathf.Sqrt(distSquared(point, lineSegPointA)), lineSegPointA);
        }else if(t > 1.0){ // Beyond the 'w' end of the segment
            return new Tuple<float, Vector2>(Mathf.Sqrt(distSquared(point, lineSegPointB)), lineSegPointB);
        }
        // Projection falls on the segment
        Vector2 closestPoint = new Vector2(lineSegPointA.x+t * (lineSegPointB.x-lineSegPointA.x), lineSegPointA.y+t * (lineSegPointB.y-lineSegPointA.y));
        return new Tuple<float, Vector2>(Mathf.Sqrt(distSquared(point, closestPoint)), closestPoint);
    }

    /** Find and return all closed polygons of length four given a list of line segments.
     *
     *  The returned points are in order.
     *
     *  Useful for finding walls within Unity that are closed (e.g. boundary box, cover). 
     *
     *  This is ugly and slow (i.e. O(N^4) I think), but it works for what I needed and there typically aren't many
     *  walls in AutoTank.
     */
    public static List<Quad<Vector2, Vector2, Vector2, Vector2>> FindAllClosedPolygonsofLengthFour(List<Tuple<Vector2, Vector2>> allWalls){
        Dictionary<string, Quad<Vector2, Vector2, Vector2, Vector2>> candidates = new Dictionary<string, Quad<Vector2, Vector2, Vector2, Vector2>>();
        for(int i=0; i<allWalls.Count; i++){
            Tuple<Vector2, Vector2> wallX = allWalls[i];
            Vector2 pointA = wallX.first;
            Vector2 pointB = wallX.second;

            for(int ii=0; ii<allWalls.Count; ii++){
                if(i==ii){continue;} // this is the same wall that's already in our segment
                Tuple<Vector2, Vector2> wallY = allWalls[ii];
                Vector2 pointC;
                if(wallY.first == pointB){pointC = wallY.second;}
                else if(wallY.second == pointB){pointC = wallY.first;}
                else{continue;} // these two walls don't connect

                for(int iii=0; iii<allWalls.Count; iii++){
                    if(i==iii || ii==iii){continue;} // this wall is already in our segment
                    Tuple<Vector2, Vector2> wallZ = allWalls[iii];
                    Vector2 pointD;
                    if(wallZ.first == pointC){pointD = wallZ.second;}
                    else if (wallZ.second == pointC){pointD = wallZ.first;}
                    else{continue;} // these two walls don't connect
                    
                    for(int iiii=0; iiii<allWalls.Count; iiii++){
                        if(i==iiii || ii==iiii || iii==iiii){continue;} // this wall is already in our segment
                        Tuple<Vector2, Vector2> wallW = allWalls[iiii];
                        if((wallW.first == pointA && wallW.second == pointD) || (wallW.first == pointD && wallW.second == pointA)){}
                        else{continue;} // This segment does NOT complete our segment, so continue looking

                        // We need to provide a unique KEY for each set of walls that doesn't depend on order, so we
                        // sort the order indices numerically and make a string out them. 
                        List<int> wallList = new List<int>();
                        wallList.Add(i);
                        wallList.Add(ii);
                        wallList.Add(iii);
                        wallList.Add(iiii);
                        wallList.Sort();
                        string key = String.Format("{0} {1} {2} {3}", wallList[0], wallList[1], wallList[2], wallList[3]);
                        
                        if(candidates.ContainsKey(key)){continue;} // This box is already in the candidates dict
                        candidates[key] = new Quad<Vector2, Vector2, Vector2, Vector2>(pointA, pointB, pointC, pointD);                    
                    }
                }
            }
        }

        List<Quad<Vector2, Vector2, Vector2, Vector2>> results = new List<Quad<Vector2, Vector2, Vector2, Vector2>>();
        foreach (KeyValuePair<string, Quad<Vector2, Vector2, Vector2, Vector2>> points in candidates) {
            results.Add(points.Value);
        }
        return results;
    }

    /** Assuming the boundary walls form a rectangle, return the largest rectangle's points (by area) and Rect.
     *
     *  We pass both because in some cases points are easier and in other cases the (Unity) Rect is easier.
     *
     *  Do NOT ask for this each tick. Save it once but check for it in genAction like so: 
     *              
     *      public class ExampleBehavior : AT.UBF.Behavior {
     *          private Quad<Vector2, Vector2, Vector2, Vector2> boundaryPoints;
     *          private Rect boundaryRect;
     *
     *          protected override AT.UBF.Action genActionInternal(AT.UBF.State state){
     *              if(boundaryPoints == null){ // will be null when we just loaded the game and hit play
     *                  Tuple<Quad<Vector2, Vector2, Vector2, Vector2>, Rect> bounds = Util.FindBoundaryWalls(state.walls); 
     *                  boundaryPoints = bounds.first;
     *                  boundaryRect = bounds.second;
     *              }
     *              
     *              // Do something with the boundary info
     *              AT.UBF.Action action = new AT.UBF.Action();
     *              return action;
     *          }
     *      }
     */
    public static Tuple<Quad<Vector2, Vector2, Vector2, Vector2>, Rect> FindBoundaryWalls(List<Tuple<Vector2, Vector2>> allWalls){
        List<Quad<Vector2, Vector2, Vector2, Vector2>> polygons = FindAllClosedPolygonsofLengthFour(allWalls);
        float maxArea = 0.0f;
        int largestArea = 0;
        Rect largestRect = new Rect();

        for(int i=0; i<polygons.Count; i++){
            Quad<Vector2, Vector2, Vector2, Vector2> polygon = polygons[i];
            Vector2 pA = polygon.first;
            Vector2 pB = polygon.second;
            Vector2 pC = polygon.third;
            Vector2 pD = polygon.fourth;

            float minX = pA.x;
            float minY = pA.y;
            float maxX = pA.x;
            float maxY = pA.y;

            if(pB.x < minX){minX = pB.x;}
            if(pB.x > maxX){maxX = pB.x;}
            if(pC.x < minX){minX = pC.x;}
            if(pC.x > maxX){maxX = pC.x;}
            if(pD.x < minX){minX = pD.x;}
            if(pD.x > maxX){maxX = pD.x;}

            if(pB.y < minY){minY = pB.y;}
            if(pB.y > maxY){maxY = pB.y;}
            if(pC.y < minY){minY = pC.y;}
            if(pC.y > maxY){maxY = pC.y;}
            if(pD.y < minY){minY = pD.y;}
            if(pD.y > maxY){maxY = pD.y;}

            float xLen = Mathf.Abs(minX - maxX);
            float yLen = Mathf.Abs(minY - maxY);

            float area = xLen * yLen;
            if(area > maxArea){
                largestArea = i;
                maxArea = area;
                largestRect = Rect.MinMaxRect(minX, minY, maxX, maxY);
            }
        }

        return new Tuple<Quad<Vector2, Vector2, Vector2, Vector2>, Rect>(polygons[largestArea], largestRect);
    }

    /** Helper method for loading a float argument from a JSON dictionary.
     *
     *  Because we have to do a double cast for floats returned from MiniJSON, and handle default values AND give good
     *  error messages, this boilerplate code quickly overwhelmed several Behaviors. I pulled it out to this method to
     *  hopefully ease the pain.
     *
     *  TODO: Should we just pass in the Behavior object as the second arg and get do behavior.GetType().Name ourselves?
     */
    public static float FloatFromJsonDict(string keyName, float defaultValue, string behaviorName, Dictionary<string, object> dict){
        float result = defaultValue;
        try{
            // MiniJSON returns floats as doubles
            result = (float) ((double) dict[keyName]); 
        }catch(KeyNotFoundException){
            Debug.Log(String.Format("'{0}' key not found in {1} Arguments. Using default of {2}.", keyName, behaviorName, defaultValue));
        }catch(InvalidCastException){
            Debug.LogError(String.Format("'{0}' value in {1} Arguments is not a valid float", keyName, behaviorName));
        }
        return result;
    }
}